# Generated by Django 3.1.5 on 2021-03-07 12:55

import secrets

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("core", "0004_auto_20210305_1855"),
    ]

    operations = [
        migrations.CreateModel(
            name="GuestUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("email", models.EmailField(max_length=254, verbose_name="email address")),
                ("first_name", models.CharField(max_length=254, verbose_name="first name")),
                ("last_name", models.CharField(max_length=254, verbose_name="last name")),
                ("date_of_birth", models.DateField(verbose_name="date of birth")),
                (
                    "phone",
                    models.CharField(blank=True, max_length=254, verbose_name="phone number"),
                ),
                (
                    "access_token",
                    models.CharField(default=secrets.token_urlsafe, max_length=254, unique=True),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "event",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="core.event"),
                ),
                ("qualifications", models.ManyToManyField(to="core.Qualification")),
            ],
            options={
                "unique_together": {("event", "email", "first_name", "last_name")},
            },
        ),
        migrations.CreateModel(
            name="GuestParticipation",
            fields=[
                (
                    "abstractparticipation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.abstractparticipation",
                    ),
                ),
                (
                    "guest_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="guests.guestuser"
                    ),
                ),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("core.abstractparticipation",),
        ),
        migrations.CreateModel(
            name="EventGuestShare",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "token",
                    models.CharField(default=secrets.token_urlsafe, max_length=254, unique=True),
                ),
                ("active", models.BooleanField(default=False)),
                (
                    "event",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="guest_share",
                        to="core.event",
                    ),
                ),
            ],
        ),
    ]



import datetime

from django import template
from django.template.defaultfilters import floatformat

register = template.Library()


@register.filter(name="get")
def dict_get(dictionary, key):
    return dictionary.get(key)


@register.filter(name="sum")
def _sum(elements):
    return sum(elements)


@register.filter(name="getattr")
def _getattr(obj, key):
    return getattr(obj, str(key))


@register.filter(name="timedelta_in_hours")
def timedelta_in_hours(delta):
    return (
        floatformat(delta / datetime.timedelta(hours=1), 2)
        if isinstance(delta, datetime.timedelta)
        else delta
    )



from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML
from django import forms
from django.utils.translation import gettext_lazy as _
from django_select2.forms import Select2Widget

from ephios.core.forms.events import BaseEventPluginFormMixin
from ephios.extra.widgets import CustomDateInput
from ephios.plugins.eventautoqualification.models import EventAutoQualificationConfiguration


class EventAutoQualificationForm(BaseEventPluginFormMixin, forms.ModelForm):
    class Meta:
        model = EventAutoQualificationConfiguration
        fields = ["qualification", "expiration_date", "mode", "extend_only", "needs_confirmation"]
        widgets = {
            "qualification": Select2Widget,
            "expiration_date": CustomDateInput,
        }

    def __init__(self, *args, event=None, edit_permission=False, **kwargs):
        kwargs.setdefault("prefix", "autoqualification")
        self.event = event
        try:
            kwargs.setdefault(
                "instance", EventAutoQualificationConfiguration.objects.get(event=self.event)
            )
        except EventAutoQualificationConfiguration.DoesNotExist:
            pass

        super().__init__(*args, **kwargs)

        self.fields["qualification"].required = False
        self.helper = FormHelper(self)
        if not edit_permission:
            self.helper.layout.insert(
                0,
                HTML(
                    "<p>"
                    + str(
                        _(
                            "You don't have permission to grant qualifications, so you can't edit these settings."
                        )
                    )
                    + "</p>"
                ),
            )
            for field in self.fields.values():
                field.disabled = True

    def save(self, commit=True):
        if self.cleaned_data.get("qualification"):
            self.instance.event = self.event
            super().save(commit)
        elif self.instance.pk:
            self.instance.delete()

    @property
    def heading(self):
        return _("Automatic qualification")

    def is_function_active(self):
        return bool(self.instance.pk)



from django.dispatch import receiver

from ephios.core.signals import event_forms, participation_finished
from ephios.plugins.eventautoqualification.consequences import create_qualification_consequence
from ephios.plugins.eventautoqualification.forms import EventAutoQualificationForm


@receiver(
    event_forms,
    dispatch_uid="ephios.plugins.eventautoqualification.signals.configuration_form",
)
def configuration_form(sender, event, request, **kwargs):
    return [
        EventAutoQualificationForm(
            request.POST or None,
            event=event,
            edit_permission=request.user.has_perm("core.change_userprofile"),
        )
    ]


participation_finished.connect(
    create_qualification_consequence,
    dispatch_uid="ephios.plugins.eventautoqualification.signals.create_qualification_consequence",
)



from django.utils.translation import gettext_lazy as _

from ephios.core.models import AbstractParticipation
from ephios.plugins.basesignup.signup.common import QualificationMinMaxBaseSignupMethod


class RequestConfirmSignupMethod(QualificationMinMaxBaseSignupMethod):
    slug = "request_confirm"
    verbose_name = _("Request and confirm")
    description = _(
        """This method lets people request participation. Responsibles can then confirm the participation."""
    )
    registration_button_text = _("Request")
    signup_success_message = _("You have successfully requested a participation for {shift}.")
    signup_error_message = _("Requesting a participation failed: {error}")

    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        participation.state = AbstractParticipation.States.REQUESTED
        return participation



from django.utils.translation import gettext_lazy as _

from ephios.core.plugins import PluginConfig


class PluginApp(PluginConfig):
    name = "ephios.plugins.guests"

    class EphiosPluginMeta:
        name = _("Guest Participations")
        author = "Ephios Team"
        description = _(
            "This plugins allows you to accept signups for individual events from people without an account, using a public signup form."
        )

    def ready(self):
        from . import signals  # pylint: disable=unused-import



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import QuerySet
from django.db.models.functions import Cast
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView
from django_select2.forms import Select2Widget

from ephios.core.models import UserProfile
from ephios.core.widgets import UserProfileWidget
from ephios.extra.mixins import CustomPermissionRequiredMixin
from ephios.extra.widgets import CustomDateInput
from ephios.modellogging.models import LogEntry


class LogFilterForm(forms.Form):
    object_type = forms.ChoiceField(
        choices=[("", "---------")],
        required=False,
        label=_("Concerns"),
        widget=Select2Widget,
        error_messages={
            "invalid_choice": _("%(value)s does not have any associated entries."),
        },
    )
    object_id = forms.IntegerField(required=False, widget=forms.HiddenInput)
    user = forms.ModelChoiceField(
        queryset=UserProfile.objects.all(),
        widget=UserProfileWidget,
        required=False,
        label=_("Acting User"),
    )
    date = forms.DateField(widget=CustomDateInput, required=False, label=_("Date"))
    search = forms.CharField(
        required=False,
        label=_("Contents"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.content_types = {}
        qs: QuerySet[ContentType] = ContentType.objects.filter(
            pk__in=LogEntry.objects.values_list("attached_to_object_type", flat=True)
        )
        for content_type in qs:
            key = content_type.model.lower()
            self.fields["object_type"].choices.append(
                (key, content_type.model_class()._meta.verbose_name_plural)
            )
            self.content_types[key] = content_type
        self.fields["object_type"].widget.choices = self.fields["object_type"].choices

    def filter(self, queryset):
        if not self.is_valid():
            return queryset

        kw = {}
        if object_type := self.cleaned_data.get("object_type"):
            kw["attached_to_object_type"] = self.content_types[object_type]
        if object_id := self.cleaned_data.get("object_id"):
            kw["attached_to_object_id"] = object_id
        if user := self.cleaned_data.get("user"):
            kw["user"] = user
        if date := self.cleaned_data.get("date"):
            kw["datetime__date"] = date

        queryset = queryset.filter(**kw)

        if search := self.cleaned_data.get("search"):
            queryset = queryset.annotate(
                data_string=Cast("data", output_field=models.TextField()),
            ).filter(data_string__icontains=search)

        return queryset

    @cached_property
    def content_object(self):
        if not self.is_valid():
            return None
        if (object_type := self.cleaned_data.get("object_type")) and (
            object_id := self.cleaned_data.get("object_id")
        ):
            return self.content_types[object_type].get_object_for_this_type(id=object_id)


class LogView(CustomPermissionRequiredMixin, ListView):
    template_name = "core/logentry_list.html"
    model = LogEntry
    permission_required = "modellogging.view_logentry"
    paginate_by = 20

    @cached_property
    def filter_form(self):
        return LogFilterForm(self.request.GET or None)

    def get_context_data(self, **kwargs):
        return super().get_context_data(filter_form=self.filter_form, **kwargs)

    def get_queryset(self):
        return self.filter_form.filter(super().get_queryset())



# Generated by Django 3.2.7 on 2021-09-03 12:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0011_auto_20210903_1258"),
    ]

    operations = [
        migrations.RenameField(
            model_name="qualification",
            old_name="included_qualifications",
            new_name="includes",
        ),
    ]



from django import template

from ephios.extra.bootstrap import render_alert

register = template.Library()


@register.simple_tag(name="render_alert")
def render_alert_tag(content, alert_type="info", dismissible=True):
    return render_alert(content, alert_type, dismissible)


@register.filter(name="formset_errors")
def formset_errors(formset):
    if errors := formset.non_form_errors():
        return render_alert("<br>".join(errors), alert_type="danger")
    return ""



import operator
from datetime import date, datetime

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db.models import IntegerField, OuterRef, Q, Subquery
from django.db.models.fields.json import KeyTransform
from django.db.models.functions import Cast
from django.template.defaultfilters import floatformat
from django.utils.formats import date_format
from django.utils.translation import gettext_lazy as _
from guardian.shortcuts import get_objects_for_user

from ephios.core.models import (
    Consequence,
    Event,
    Qualification,
    QualificationGrant,
    Shift,
    UserProfile,
    WorkingHours,
)
from ephios.core.signals import register_consequence_handlers


def installed_consequence_handlers():
    for _, handlers in register_consequence_handlers.send_to_all_plugins(None):
        yield from (h() for h in handlers)


def consequence_handler_from_slug(slug):
    for handler in installed_consequence_handlers():
        if handler.slug == slug:
            return handler
    raise ValueError(_("Consequence Handler '{slug}' was not found.").format(slug=slug))


def editable_consequences(user):
    handlers = list(installed_consequence_handlers())
    qs = Consequence.objects.all().select_related("user")
    for handler in handlers:
        qs = handler.filter_queryset(qs, user)
    return qs.filter(slug__in=map(operator.attrgetter("slug"), handlers)).distinct()


def pending_consequences(user):
    qs = Consequence.objects.filter(user=user, state=Consequence.States.NEEDS_CONFIRMATION)
    return qs


class ConsequenceError(Exception):
    pass


class BaseConsequenceHandler:
    @property
    def slug(self):
        raise NotImplementedError

    @classmethod
    def execute(cls, consequence):
        """
        Gets a consequence and tries to execute whatever it is the consequence wants to happen.
        """
        raise NotImplementedError

    @classmethod
    def render(cls, consequence):
        """
        Return html describing the action to be done as a consequence of what.
        Return None if you cannot handle this consequence.
        """
        raise NotImplementedError

    @classmethod
    def filter_queryset(cls, qs, user: UserProfile):
        """
        Return a filtered that excludes consequences with the slug of this class that the user is not allowed to edit.
        Consequences should also be annotated with values needed for rendering.
        """
        raise NotImplementedError


class WorkingHoursConsequenceHandler(BaseConsequenceHandler):
    slug = "ephios.grant_working_hours"

    @classmethod
    def create(
        cls,
        user: UserProfile,
        when: date,
        hours: float,
        reason: str,
    ):
        return Consequence.objects.create(
            slug=cls.slug,
            user=user,
            data=dict(hours=hours, date=when, reason=reason),
        )

    @classmethod
    def execute(cls, consequence):
        WorkingHours.objects.create(
            user=consequence.user,
            date=consequence.data["date"],
            hours=consequence.data["hours"],
            reason=consequence.data.get("reason"),
        )

    @classmethod
    def render(cls, consequence):
        return _("{user} obtains {hours} working hours for {reason} on {date}").format(
            user=consequence.user.get_full_name(),
            hours=floatformat(consequence.data.get("hours"), arg=-2),
            reason=consequence.data.get("reason"),
            date=date_format(consequence.data.get("date")),
        )

    @classmethod
    def filter_queryset(cls, qs, user: UserProfile):
        return qs.filter(
            ~Q(slug=cls.slug)
            | Q(
                user__groups__in=get_objects_for_user(
                    user, "decide_workinghours_for_group", klass=Group
                )
            )
        )


class QualificationConsequenceHandler(BaseConsequenceHandler):
    slug = "ephios.grant_qualification"

    @classmethod
    def create(
        cls,
        user: UserProfile,
        qualification: Qualification,
        expires: datetime = None,
        shift: Shift = None,
    ):
        return Consequence.objects.create(
            slug=cls.slug,
            user=user,
            data=dict(
                qualification_id=qualification.id,
                event_id=None if shift is None else shift.event_id,
                expires=expires,
            ),
        )

    @classmethod
    def execute(cls, consequence):
        qg, created = QualificationGrant.objects.get_or_create(
            defaults=dict(
                expires=consequence.data["expires"],
            ),
            user=consequence.user,
            qualification_id=consequence.data["qualification_id"],
        )
        if not created:
            qg.expires = max(
                qg.expires, consequence.data["expires"], key=lambda dt: dt or datetime.max
            )
            qg.save()

    @classmethod
    def render(cls, consequence):
        # Get all the strings we need from the annotations, or fetch them from DB as backup
        try:  # try the annotation
            event_title = consequence.event_title
        except AttributeError:
            if event_id := consequence.data["event_id"]:  # fetch from DB as backup
                event_title = Event.objects.get(id=event_id).title
            else:  # no event has been associated
                event_title = None

        try:
            qualification_title = consequence.qualification_title
        except AttributeError:
            qualification_title = Qualification.objects.get(
                id=Cast(consequence.data["qualification_id"], IntegerField())
            ).title

        if expires := consequence.data.get("expires"):
            expires = date_format(expires)

        user = consequence.user.get_full_name()

        # build string based on available data

        if event_title:
            s = _("{user} acquires '{qualification}' after participating in {event}.").format(
                user=user, qualification=qualification_title, event=event_title
            )
        else:
            s = _("{user} acquires '{qualification}'.").format(
                user=user,
                qualification=qualification_title,
            )

        if expires:
            s += " " + _("(valid until {expires_str})").format(expires_str=expires)
        return s

    @classmethod
    def filter_queryset(cls, qs, user: UserProfile):
        qs = qs.annotate(
            qualification_id=Cast(KeyTransform("qualification_id", "data"), IntegerField()),
            event_id=Cast(KeyTransform("event_id", "data"), IntegerField()),
        ).annotate(
            qualification_title=Subquery(
                Qualification.objects.filter(id=OuterRef("qualification_id")).values("title")[:1]
            ),
            event_title=Subquery(Event.objects.filter(id=OuterRef("event_id")).values("title")[:1]),
        )

        return qs.filter(
            ~Q(slug=cls.slug)
            # Qualifications can be granted by people who...
            | Q(  # are responsible for the event the consequence originated from, if applicable
                event_id__in=get_objects_for_user(user, perms="change_event", klass=Event),
            )
            | Q(  # can edit the affected user anyway
                user__in=get_objects_for_user(
                    user, perms="change_userprofile", klass=get_user_model()
                )
            )
        )



import collections.abc
import json

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import QuerySet
from django.utils import dateparse
from django.utils.functional import SimpleLazyObject

# pylint: disable=protected-access


class LogJSONEncoder(DjangoJSONEncoder):
    """
    Encoder designed to handle querysets and model instances while falling back to their string representation in the corresponding Decoder.
    """

    def default(self, o):
        queryset_like = SimpleLazyObject(
            lambda: isinstance(o, collections.abc.Collection)
            and all(isinstance(elem, models.Model) for elem in o)
            and len({type(elem) for elem in o}) == 1
        )
        if isinstance(o, QuerySet) or queryset_like:
            pks, strs = [], []
            for instance in o:
                pks.append(instance.pk)
                strs.append(str(instance))
            return {
                "__model__": "__queryset__",
                "pks": pks,
                "strs": strs,
                "contenttype_id": ContentType.objects.get_for_model(
                    getattr(o, "model", None) or type(next(iter(o)))
                ).id,
            }
        if o == set():
            return []
        if isinstance(o, models.Model):
            return {
                "__model__": "__instance__",
                "pk": o.pk,
                "str": str(o),
                "contenttype_id": ContentType.objects.get_for_model(o).id,
            }
        return super().default(o)


class LogJSONDecoder(json.JSONDecoder):
    """
    Decoder designed to handle querysets and model instances while falling back to their string representation from the corresponding Encoder.
    """

    def __init__(self, *args, **kargs):
        super().__init__(*args, object_hook=self.custom_hook, **kargs)

    def custom_hook(self, d):
        if d.get("__model__") == "__queryset__":
            Model = ContentType.objects.get_for_id(d["contenttype_id"]).model_class()
            objects = {obj.pk: obj for obj in Model._base_manager.filter(pk__in=d["pks"])}
            return [objects.get(pk, s) for pk, s in zip(d["pks"], d["strs"])]
        if d.get("__model__") == "__instance__":
            try:
                return ContentType.objects.get_for_id(d["contenttype_id"]).get_object_for_this_type(
                    pk=d["pk"]
                )
            except ObjectDoesNotExist:
                return d["str"]
        for k, v in d.items():
            if isinstance(v, str):
                if (dt := dateparse.parse_datetime(v)) is not None:
                    d[k] = dt
                if (dt := dateparse.parse_date(v)) is not None:
                    d[k] = dt
        return d



import itertools
import urllib
from collections import defaultdict
from typing import Dict
from urllib.error import URLError

from django.db import transaction
from django.db.models import Count
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.exceptions import ValidationError as RestValidationError
from rest_framework.parsers import JSONParser

from ephios.core.models import Qualification, QualificationCategory
from ephios.plugins.qualification_management.serializers import QualificationFixtureSerializer


class DeserializedQualification:
    def __init__(self, validated_data):
        self.m2m_data = {
            "includes": validated_data["includes"],
            "included_by": validated_data["included_by"],
        }
        self.object = Qualification(
            **{key: validated_data[key] for key in ("title", "abbreviation", "uuid")},
        )
        self.category = QualificationCategory(
            **{key: validated_data["category"][key] for key in ("title", "uuid")},
        )

    def __hash__(self):
        return hash(self.object.uuid)


class RepoError(Exception):
    pass


def fetch_deserialized_qualifications_from_repo():
    repo_urls = (
        stripped_url
        for url in global_preferences_registry.manager()
        .get("general__qualification_management_repos")
        .splitlines()
        if (stripped_url := url.strip())
    )
    try:
        for repo_url in repo_urls:
            with urllib.request.urlopen(repo_url) as request:
                data = JSONParser().parse(request)
                serializer = QualificationFixtureSerializer(data=data, many=True)
                assert serializer.is_valid(raise_exception=True)
                yield from (DeserializedQualification(d) for d in serializer.validated_data)
    except (URLError, RestValidationError) as e:
        raise RepoError from e


class QualificationGraph:
    """Class used to compute the inclusions used in the qualification graph."""

    def __init__(self):
        # inclusions is a dict where keys are nodes and values are lists of nodes the key node has an edge of inclusion to
        self.inclusions = defaultdict(set)

    def add_qualification(self, uuid, inclusions):
        """Add a new qualification to the graph."""
        self.inclusions[uuid] |= inclusions

    def parents(self, child):
        for parent, children in self.inclusions.items():
            if child in children:
                yield parent

    def remove_qualification(self, uuid):
        """Remove a qualification, keeping the inclusions intact."""
        children = self.inclusions[uuid]
        for parent in self.parents(uuid):
            self.inclusions[parent] |= children
            self.inclusions[parent].remove(uuid)
        del self.inclusions[uuid]


class QualificationChangeManager:
    def __init__(self):
        self.deserialized_qualifications_to_add = set()
        self.inclusion_supporting_deserialized_qualifications = set()
        self.qualifications_to_delete_fixing_inclusion = set()

        self.existing_qualifications_by_uuid: Dict[str, Qualification] = {
            str(obj.uuid): obj for obj in Qualification.objects.all()
        }

    def add_deserialized_qualifications_to_db(self, *deserialized_qualifications):
        self.deserialized_qualifications_to_add |= set(deserialized_qualifications)
        return self

    def add_inclusions_of_deserialized_qualifications(self, *deserialized_qualifications):
        self.inclusion_supporting_deserialized_qualifications |= set(deserialized_qualifications)
        return self

    def remove_qualifications_from_db_fixing_inclusion(self, *qualifications):
        self.qualifications_to_delete_fixing_inclusion |= set(qualifications)
        return self

    def _attach_deferred_categories_to_deserialized_qualifications(self):
        categories_by_uuid = {str(c.uuid): c for c in QualificationCategory.objects.all()}
        for deserialized_qualification in self.deserialized_qualifications_to_add:
            deserialized_qualification.object.category = categories_by_uuid[
                str(deserialized_qualification.category.uuid)
            ]

    def _build_inclusion_graph(self):
        graph = QualificationGraph()
        # add qualifications from the repo
        for deserialized_qualification in itertools.chain(
            self.deserialized_qualifications_to_add,
            self.inclusion_supporting_deserialized_qualifications,
        ):
            this_uuid = str(deserialized_qualification.object.uuid)
            inclusions = set(deserialized_qualification.m2m_data["includes"])
            graph.add_qualification(uuid=this_uuid, inclusions=inclusions)
            for included_by in deserialized_qualification.m2m_data["included_by"]:
                graph.add_qualification(uuid=str(included_by), inclusions={str(this_uuid)})

        # add existing qualifications and their inclusions
        repository_qualification_uuids = set(graph.inclusions.keys())
        for qualification in self.existing_qualifications_by_uuid.values():
            this_uuid = str(qualification.uuid)
            inclusions = {str(included.uuid) for included in qualification.includes.all()}
            if this_uuid in repository_qualification_uuids:
                # This existing qualification is also part of the repo.
                # Inclusions to other repo qualifications, that are originally not part of the repo should not be reproduced, so we don't add them.
                inclusions -= repository_qualification_uuids
            graph.add_qualification(uuid=this_uuid, inclusions=inclusions)

        # now remove supporting-only repo qualifications and explicitly removed ones, maintaining graph connectivity
        for deserialized_qualification in self.inclusion_supporting_deserialized_qualifications:
            graph.remove_qualification(str(deserialized_qualification.object.uuid))
        for qualification in self.qualifications_to_delete_fixing_inclusion:
            graph.remove_qualification(str(qualification.uuid))

        self.graph = graph

    def _perform_qualification_db_operations(self):
        # Delete the qualifications we don't need.
        Qualification.objects.filter(
            pk__in={obj.pk for obj in self.qualifications_to_delete_fixing_inclusion}
        ).delete()

        # Create not yet existing qualification objects (without m2m inclusion data)
        for deserialized_qualification in self.deserialized_qualifications_to_add:
            uuid = str(deserialized_qualification.object.uuid)
            if uuid not in self.existing_qualifications_by_uuid:
                deserialized_qualification.object.save()
                self.existing_qualifications_by_uuid[uuid] = deserialized_qualification.object

        # Finally, set m2m inclusions
        for uuid, inclusions in self.graph.inclusions.items():
            qualification = self.existing_qualifications_by_uuid[uuid]
            qualification.includes.set(
                [self.existing_qualifications_by_uuid[inclusion] for inclusion in inclusions]
            )

    def _prepare_qualification_category_db(self):
        uuids_of_used_categories = {
            str(deserialized_qualification.category.uuid)
            for deserialized_qualification in self.deserialized_qualifications_to_add
        }
        qualification_categories = {
            deserialized_qualifcation.category.uuid: deserialized_qualifcation.category
            for deserialized_qualifcation in self.deserialized_qualifications_to_add
        }
        for category in qualification_categories.values():
            if str(category.uuid) in uuids_of_used_categories:
                QualificationCategory.objects.get_or_create(
                    uuid=category.uuid,
                    defaults=dict(
                        title=category.title,
                    ),
                )

    def _delete_unneeded_qualification_categories(self):
        repo_categories_uuids = {
            str(deserialized_qualifications.category.uuid)
            for deserialized_qualifications in itertools.chain(
                self.deserialized_qualifications_to_add,
                self.inclusion_supporting_deserialized_qualifications,
            )
        }
        QualificationCategory.objects.annotate(qualification_count=Count("qualifications")).filter(
            uuid__in=repo_categories_uuids, qualification_count=0
        ).delete()

    @transaction.atomic()
    def commit(self):
        # To keep the inclusion graph intact, here's what to look out for:
        # - check that newly added qualifications are included in the correct existing ones
        # - check that newly added qualifications include the correct existing ones
        # - check that removing a qualification does not break a chain of inclusion
        # - check that adding a subset of a fixture's qualifications keeps all inclusions

        self.inclusion_supporting_deserialized_qualifications -= (
            self.deserialized_qualifications_to_add
        )

        self._prepare_qualification_category_db()
        self._attach_deferred_categories_to_deserialized_qualifications()
        self._build_inclusion_graph()
        self._perform_qualification_db_operations()
        self._delete_unneeded_qualification_categories()



from typing import List
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.auth.tokens import default_token_generator
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.formats import date_format
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import gettext_lazy as _
from guardian.shortcuts import get_users_with_perms

from ephios.core.models import AbstractParticipation, Event, LocalParticipation, UserProfile
from ephios.core.models.users import Consequence, Notification
from ephios.core.signals import register_notification_types
from ephios.core.signup.participants import LocalUserParticipant


def installed_notification_types():
    for _, handlers in register_notification_types.send_to_all_plugins(None):
        yield from (h for h in handlers)


def enabled_notification_types():
    for _, handlers in register_notification_types.send(None):
        yield from (h for h in handlers)


def notification_type_from_slug(slug):
    for notification in installed_notification_types():
        if notification.slug == slug:
            return notification
    raise ValueError(_("Notification type '{slug}' was not found.").format(slug=slug))


class AbstractNotificationHandler:
    unsubscribe_allowed = True

    @property
    def slug(self):
        return NotImplementedError

    @property
    def description(self):
        return NotImplementedError

    @classmethod
    def get_subject(cls, notification):
        raise NotImplementedError

    @classmethod
    def as_plaintext(cls, notification):
        raise NotImplementedError

    @classmethod
    def as_html(cls, notification):
        return render_to_string("email_base.html", {"message_text": cls.as_plaintext(notification)})

    @classmethod
    def get_url(cls, notification):
        return None


class ProfileUpdateNotification(AbstractNotificationHandler):
    slug = "ephios_profile_update"
    title = _("Your profile has been edited")

    @classmethod
    def send(cls, user: UserProfile, **kwargs):
        return Notification.objects.create(slug=cls.slug, user=user, data=kwargs)

    @classmethod
    def get_subject(cls, notification):
        return _("ephios account updated")

    @classmethod
    def as_plaintext(cls, notification):
        return _(
            "You're receiving this email because your account at ephios has been updated.\n"
            "You can see the changes in your profile: {url}\n"
            "Your username is your email address: {email}\n"
        ).format(url=notification.get_url(), email=notification.user.email)

    @classmethod
    def get_url(cls, notification):
        return urljoin(settings.GET_SITE_URL(), reverse("core:profile"))


class NewProfileNotification(AbstractNotificationHandler):
    slug = "ephios_new_profile"
    title = _("A new profile has been created")
    unsubscribe_allowed = False

    @classmethod
    def send(cls, user: UserProfile, **kwargs):
        uid = urlsafe_base64_encode(force_bytes(user.id))
        token = default_token_generator.make_token(user)
        return Notification.objects.create(
            slug=cls.slug, user=user, data={**{"uidb64": uid, "token": token}, **kwargs}
        )

    @classmethod
    def get_subject(cls, notification):
        return _("Welcome to ephios!")

    @classmethod
    def as_plaintext(cls, notification):
        reset_link = reverse("password_reset_confirm", kwargs=notification.data)
        return _(
            "You're receiving this email because a new account has been created for you at ephios.\n"
            "Please go to the following page and choose a password: {url}\n"
            "Your username is your email address: {email}\n"
        ).format(url=urljoin(settings.GET_SITE_URL(), reset_link), email=notification.user.email)

    @classmethod
    def as_html(cls, notification):
        return render_to_string(
            "core/mails/new_account_email.html",
            {
                **{"site_url": settings.GET_SITE_URL(), "email": notification.user.email},
                **notification.data,
            },
        )


class NewEventNotification(AbstractNotificationHandler):
    slug = "ephios_new_event"
    title = _("A new event has been added")

    @classmethod
    def send(cls, event: Event, **kwargs):
        notifications = []
        for user in get_users_with_perms(event, only_with_perms_in=["view_event"]):
            notifications.append(
                Notification(slug=cls.slug, user=user, data=dict(event_id=event.id, **kwargs))
            )
        Notification.objects.bulk_create(notifications)

    @classmethod
    def get_subject(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return _("New {type}: {title}").format(type=event.type, title=event.title)

    @classmethod
    def as_plaintext(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return _(
            "A new {type} ({title}, {location}) has been added.\n"
            "Further information: {description}\n"
            "You can view the event here: {url}"
        ).format(
            type=event.type,
            title=event.title,
            location=event.location,
            description=event.description,
            url=notification.get_url(),
        )

    @classmethod
    def as_html(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return render_to_string(
            "core/mails/new_event.html", {"event": event, "site_url": settings.GET_SITE_URL()}
        )

    @classmethod
    def get_url(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return urljoin(settings.GET_SITE_URL(), event.get_absolute_url())


class ParticipationMixin:
    @classmethod
    def get_url(cls, notification):
        shift = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift
        return urljoin(settings.GET_SITE_URL(), shift.get_absolute_url())

    @classmethod
    def send(cls, participation: AbstractParticipation):
        user = (
            participation.user
            if participation.get_real_instance_class() == LocalParticipation
            else None
        )
        Notification.objects.create(
            slug=cls.slug,
            user=user,
            data=dict(participation_id=participation.id, email=participation.participant.email),
        )


class ParticipationConfirmedNotification(ParticipationMixin, AbstractNotificationHandler):
    slug = "ephios_participation_confirmed"
    title = _("Your participation has been confirmed")

    @classmethod
    def get_subject(cls, notification):
        event = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift.event
        return _("Participation confirmed for {event}").format(event=event)

    @classmethod
    def as_plaintext(cls, notification):
        participation: AbstractParticipation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        message = _("Your participation for {shift} is now confirmed.").format(
            shift=participation.shift
        )
        if participation.has_customized_signup():
            message += f'\n\n{_("Your time is")} {participation.get_time_display()}'
        return message


class ParticipationRejectedNotification(ParticipationMixin, AbstractNotificationHandler):
    slug = "ephios_participation_rejected"
    title = _("Your participation has been rejected")

    @classmethod
    def get_subject(cls, notification):
        event = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift.event
        return _("Participation rejected for {event}").format(event=event)

    @classmethod
    def as_plaintext(cls, notification):
        shift = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift
        return _("Your participation for {shift} has been rejected by a responsible user.").format(
            shift=shift
        )


class ParticipationCustomizationNotification(ParticipationMixin, AbstractNotificationHandler):
    slug = "ephios_participation_customized"
    title = _("A confirmed participation of yours has been tweaked by a responsible")

    @classmethod
    def send(
        cls, participation: AbstractParticipation, claims: List[str]
    ):  # pylint: disable=arguments-differ
        user = (
            participation.user
            if participation.get_real_instance_class() == LocalParticipation
            else None
        )
        Notification.objects.create(
            slug=cls.slug,
            user=user,
            data=dict(
                participation_id=participation.id,
                email=participation.participant.email,
                claims=claims,
            ),
        )

    @classmethod
    def get_subject(cls, notification):
        shift = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift
        return _("Participation tweaked for {shift}").format(shift=shift)

    @classmethod
    def as_plaintext(cls, notification):
        shift = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        ).shift
        message = _(
            "Your participation for {shift} has been tweaked by a responsible user."
        ).format(shift=shift)
        message += "\n\n" + "\n".join(f"- {claim}" for claim in notification.data.get("claims", []))
        return message


class ResponsibleMixin:
    def __init__(self, participation: AbstractParticipation):
        self.participation = participation

    def responsible_users(self):
        users = get_users_with_perms(
            self.participation.shift.event, only_with_perms_in=["change_event"]
        ).distinct()
        for user in users:
            if (
                self.participation.get_real_instance_class() != LocalParticipation
                or user != self.participation.user
            ):
                yield user

    def get_data(self, **kwargs):
        kwargs["disposition_url"] = urljoin(
            settings.GET_SITE_URL(),
            reverse("core:shift_disposition", kwargs=dict(pk=self.participation.shift.pk)),
        )
        kwargs["participation_id"] = self.participation.id
        return kwargs

    def send(self):
        Notification.objects.bulk_create(
            [
                Notification(
                    slug=self.slug,
                    user=user,
                    data=self.get_data(),
                )
                for user in self.responsible_users()
            ]
        )

    @classmethod
    def get_url(cls, notification):
        return notification.data.get("disposition_url")

    @classmethod
    def get_subject(cls, notification):
        participation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        participation_state = participation.get_state_display()
        return _("Participation {state} for {event}").format(
            state=participation_state, event=participation.shift.event
        )


class ResponsibleParticipationRequestedNotification(ResponsibleMixin, AbstractNotificationHandler):
    slug = "ephios_participation_responsible_requested"
    title = _("A participation has been requested for your event")

    @classmethod
    def as_plaintext(cls, notification):
        participation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        if participation.shift.signup_method.uses_requested_state:
            return _(
                "{participant} has requested a participation for {shift}. You can decide about it at {disposition_url}"
            ).format(
                shift=participation.shift,
                participant=participation.participant,
                disposition_url=notification.data.get("disposition_url"),
            )
        return _("{participant} signed up for {shift}").format(
            participant=participation.participant, shift=participation.shift
        )


class ResponsibleConfirmedParticipationDeclinedNotification(
    ResponsibleMixin, AbstractNotificationHandler
):
    slug = "ephios_participation_responsible_confirmed_declined"
    title = _("A participant declined after having been confirmed for your event")

    @classmethod
    def as_plaintext(cls, notification):
        participation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        return _("{participant} declined their participation in {shift}.").format(
            participant=participation.participant, shift=participation.shift
        )


class ResponsibleConfirmedParticipationCustomizedNotification(
    ResponsibleMixin, AbstractNotificationHandler
):
    slug = "ephios_participation_responsible_confirmed_customized"
    title = _("A confirmed participant altered their participation")

    @classmethod
    def get_subject(cls, notification):
        participation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        return _("Participation altered for {event}").format(event=participation.shift.event)

    def __init__(self, participation: AbstractParticipation, claims: List[str]):
        super().__init__(participation)
        self.claims = claims

    def get_data(self, **kwargs):
        return super().get_data(claims=self.claims)

    @classmethod
    def as_plaintext(cls, notification):
        participation = AbstractParticipation.objects.get(
            id=notification.data.get("participation_id")
        )
        message = _("{participant} altered their participation in {shift}.").format(
            participant=participation.participant, shift=participation.shift
        )
        message += "\n\n" + "\n".join(f"- {claim}" for claim in notification.data.get("claims", []))
        return message


class EventReminderNotification(AbstractNotificationHandler):
    slug = "ephios_event_reminder"
    title = _("An event has vacant spots")
    unsubscribe_allowed = False

    @classmethod
    def send(cls, event: Event):
        users_not_participating = UserProfile.objects.exclude(
            pk__in=AbstractParticipation.objects.filter(shift__event=event).values_list(
                "localparticipation__user", flat=True
            )
        ).filter(pk__in=get_users_with_perms(event, only_with_perms_in=["view_event"]))
        notifications = []
        for user in users_not_participating:
            notifications.append(
                Notification(slug=cls.slug, user=user, data=dict(event_id=event.id))
            )
        Notification.objects.bulk_create(notifications)

    @classmethod
    def get_subject(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return _("Help needed for {title}").format(title=event.title)

    @classmethod
    def as_plaintext(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return _(
            "Your support is needed for {title} ({start} - {end}). \nYou can view the event here: {url}"
        ).format(
            title=event.title,
            start=date_format(event.get_start_time(), "SHORT_DATETIME_FORMAT"),
            end=date_format(event.get_end_time(), "SHORT_DATETIME_FORMAT"),
            url=notification.get_url(),
        )

    @classmethod
    def get_url(cls, notification):
        event = Event.objects.get(pk=notification.data.get("event_id"))
        return urljoin(settings.GET_SITE_URL(), event.get_absolute_url())


class CustomEventParticipantNotification(AbstractNotificationHandler):
    slug = "ephios_custom_event_participant"
    title = _("Message to all participants")
    unsubscribe_allowed = False

    @classmethod
    def send(cls, event: Event, content: str):
        participants = set()
        for shift in event.shifts.all():
            participants.update(shift.get_participants())
        notifications = []
        for participant in participants:
            user = participant.user if isinstance(participant, LocalUserParticipant) else None
            notifications.append(
                Notification(
                    slug=cls.slug,
                    user=user,
                    data=dict(
                        email=participant.email,
                        event_id=event.id,
                        content=content,
                        event_title=event.title,
                    ),
                )
            )
        Notification.objects.bulk_create(notifications)

    @classmethod
    def get_subject(cls, notification):
        return _("Information for your participation at {title}").format(
            title=notification.data.get("event_title")
        )

    @classmethod
    def as_plaintext(cls, notification):
        return notification.data.get("content")


class ConsequenceApprovedNotification(AbstractNotificationHandler):
    slug = "ephios_consequence_approved"
    title = _("Your request has been approved")

    @classmethod
    def send(cls, consequence: Consequence):
        Notification.objects.create(
            slug=cls.slug, user=consequence.user, data=dict(consequence_id=consequence.id)
        )

    @classmethod
    def get_subject(cls, notification):
        return _("Your request has been approved")

    @classmethod
    def as_plaintext(cls, notification):
        consequence = Consequence.objects.get(id=notification.data.get("consequence_id"))
        return _('"{consequence}" has been approved.').format(consequence=consequence)


class ConsequenceDeniedNotification(AbstractNotificationHandler):
    slug = "ephios_consequence_denied"
    title = _("Your request has been denied")

    @classmethod
    def send(cls, consequence: Consequence):
        Notification.objects.create(
            slug=cls.slug, user=consequence.user, data=dict(consequence_id=consequence.id)
        )

    @classmethod
    def get_subject(cls, notification):
        return _("Your request has been denied")

    @classmethod
    def as_plaintext(cls, notification):
        consequence = Consequence.objects.get(id=notification.data.get("consequence_id"))
        return _('"{consequence}" has been denied.').format(consequence=consequence)


CORE_NOTIFICATION_TYPES = [
    ProfileUpdateNotification,
    NewProfileNotification,
    ParticipationRejectedNotification,
    ParticipationConfirmedNotification,
    ParticipationCustomizationNotification,
    ResponsibleParticipationRequestedNotification,
    ResponsibleConfirmedParticipationDeclinedNotification,
    ResponsibleConfirmedParticipationCustomizedNotification,
    NewEventNotification,
    EventReminderNotification,
    CustomEventParticipantNotification,
    ConsequenceApprovedNotification,
    ConsequenceDeniedNotification,
]



from django.dispatch import receiver

from ephios.core.signup.methods import register_signup_methods
from ephios.plugins.basesignup.signup.coupled_signup import CoupledSignupMethod
from ephios.plugins.basesignup.signup.instant import InstantConfirmationSignupMethod
from ephios.plugins.basesignup.signup.no_selfservice import NoSelfserviceSignupMethod
from ephios.plugins.basesignup.signup.request_confirm import RequestConfirmSignupMethod
from ephios.plugins.basesignup.signup.section_based import SectionBasedSignupMethod


@receiver(
    register_signup_methods,
    dispatch_uid="ephios.plugins.basesignup.signals.register_signup_methods",
)
def register_base_signup_methods(sender, **kwargs):
    return [
        InstantConfirmationSignupMethod,
        RequestConfirmSignupMethod,
        SectionBasedSignupMethod,
        NoSelfserviceSignupMethod,
        CoupledSignupMethod,
    ]



# Generated by Django 3.1.4 on 2021-01-06 21:19

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0004_auto_20201014_1648"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="qualificationgrant",
            unique_together={("qualification", "user")},
        ),
        migrations.CreateModel(
            name="WorkingHours",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("hours", models.DecimalField(decimal_places=2, max_digits=7)),
                ("reason", models.CharField(blank=True, default="", max_length=1024)),
                ("datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Consequence",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("slug", models.CharField(max_length=255)),
                ("data", models.JSONField(default=dict)),
                (
                    "state",
                    models.TextField(
                        choices=[
                            ("needs_confirmation", "needs confirmation"),
                            ("executed", "executed"),
                            ("failed", "failed"),
                            ("denied", "denied"),
                        ],
                        default="needs_confirmation",
                        max_length=31,
                    ),
                ),
                ("executed_at", models.DateTimeField(blank=True, null=True)),
                ("fail_reason", models.TextField(blank=True, max_length=255)),
                (
                    "decided_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="confirmed_consequences",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="confirmed by",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="affecting_consequences",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="affected user",
                    ),
                ),
            ],
        ),
    ]



from datetime import datetime

from django.contrib.auth.hashers import make_password
from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils.timezone import make_aware
from django.utils.translation import gettext as _
from guardian.shortcuts import assign_perm

from ephios.core.models import Event, EventType, Shift, UserProfile


def create_objects():
    admin_user = UserProfile(
        email="admin@localhost",
        first_name="Admin",
        last_name="Localhost",
        date_of_birth=datetime(year=1970, month=1, day=1),
    )
    admin_user.is_staff = True
    admin_user.is_superuser = True
    admin_user.password = make_password("admin")
    admin_user.save()

    from django.contrib.auth.models import Group

    volunteers = Group.objects.create(name=_("Volunteers"))
    volunteers.user_set.add(admin_user)
    volunteers.save()

    planners = Group.objects.create(name=_("Planners"))
    planners.user_set.add(admin_user)
    planners.save()

    managers = Group.objects.create(name=_("Managers"))
    managers.user_set.add(admin_user)
    managers.save()

    assign_perm("publish_event_for_group", planners, volunteers)
    assign_perm("core.add_event", planners)
    assign_perm("core.delete_event", planners)
    assign_perm("core.view_past_event", planners)
    assign_perm("core.view_userprofile", managers)
    assign_perm("core.add_userprofile", managers)
    assign_perm("core.change_userprofile", managers)
    assign_perm("core.delete_userprofile", managers)
    assign_perm("auth.view_group", managers)
    assign_perm("auth.add_group", managers)
    assign_perm("auth.change_group", managers)
    assign_perm("auth.delete_group", managers)

    service_type = EventType.objects.create(title=_("Service"))
    EventType.objects.create(title=_("Training"))

    user = UserProfile(
        email="user@localhost",
        first_name="User",
        last_name="Localhost",
        date_of_birth=datetime(year=1970, month=1, day=1),
    )
    user.password = make_password("user")
    user.save()
    volunteers.user_set.add(user)

    event = Event.objects.create(
        title="Concert Medical Service",
        description="Your contact is Lisa Example. Her Phone number is 012345678910",
        type=service_type,
        location="Town Square Gardens",
        active=True,
    )

    assign_perm("core.view_event", volunteers, event)

    Shift.objects.create(
        event=event,
        meeting_time=make_aware(datetime(2023, 6, 30, 15, 30)),
        start_time=make_aware(datetime(2023, 6, 30, 16, 0)),
        end_time=make_aware(datetime(2023, 7, 1, 1, 0)),
        signup_method_slug="instant_confirmation",
        signup_configuration=dict(
            minimum_age=18,
            signup_until=make_aware(datetime(2023, 6, 29, 8, 0)),
        ),
    )


class Command(BaseCommand):
    help = "Load some data for development"

    def handle(self, *args, **options):
        if UserProfile.objects.exists():
            self.stdout.write("WARNING! User objects already exist in your database.")
            if input("Are you sure you want to continue? (yes/no) ") != "yes":
                self.stdout.write("Aborting...")
                return
        with transaction.atomic():
            create_objects()
        self.stdout.write(self.style.SUCCESS("Done."))



# Generated by Django 3.2.9 on 2021-11-06 17:52

import django.db.models.deletion
import django.db.models.manager
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0013_auto_20211027_2203"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="event",
            options={
                "base_manager_name": "all_objects",
                "default_manager_name": "objects",
                "permissions": [("view_past_event", "Can view past events")],
                "verbose_name": "event",
                "verbose_name_plural": "events",
            },
        ),
        migrations.AlterModelManagers(
            name="event",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.RemoveField(
            model_name="eventtype",
            name="can_grant_qualification",
        ),
        migrations.AddField(
            model_name="abstractparticipation",
            name="comment",
            field=models.CharField(blank=True, max_length=255, verbose_name="Comment"),
        ),
        migrations.AddField(
            model_name="abstractparticipation",
            name="individual_end_time",
            field=models.DateTimeField(null=True, verbose_name="individual end time"),
        ),
        migrations.AddField(
            model_name="abstractparticipation",
            name="individual_start_time",
            field=models.DateTimeField(null=True, verbose_name="individual start time"),
        ),
        migrations.AddField(
            model_name="notification",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="abstractparticipation",
            name="state",
            field=models.IntegerField(
                choices=[
                    (0, "requested"),
                    (1, "confirmed"),
                    (2, "declined by user"),
                    (3, "rejected by responsible"),
                    (4, "getting dispatched"),
                ],
                default=4,
                verbose_name="state",
            ),
        ),
        migrations.AlterField(
            model_name="localparticipation",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="participations",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Participant",
            ),
        ),
    ]



from django.contrib.auth.models import Group
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django_select2.forms import Select2MultipleWidget
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import (
    PerInstancePreferenceRegistry,
    global_preferences_registry,
)
from dynamic_preferences.types import (
    ModelMultipleChoicePreference,
    MultipleChoicePreference,
    StringPreference,
)
from dynamic_preferences.users.registries import user_preferences_registry

import ephios
from ephios.core import plugins
from ephios.core.models import QualificationCategory, UserProfile
from ephios.core.services.notifications.backends import CORE_NOTIFICATION_BACKENDS
from ephios.core.services.notifications.types import CORE_NOTIFICATION_TYPES
from ephios.extra.preferences import JSONPreference


class EventTypeRegistry(PerInstancePreferenceRegistry):
    pass


event_type_preference_registry = EventTypeRegistry()

notifications_user_section = Section("notifications")
responsible_notifications_user_section = Section("responsible_notifications")
general_global_section = Section("general")


@global_preferences_registry.register
class OrganizationName(StringPreference):
    name = "organization_name"
    verbose_name = _("Organization name")
    default = ""
    section = general_global_section
    required = False


@global_preferences_registry.register
class RelevantQualificationCategories(ModelMultipleChoicePreference):
    name = "relevant_qualification_categories"
    section = general_global_section
    model = QualificationCategory
    default = QualificationCategory.objects.none()
    verbose_name = _("Relevant qualification categories (for user list and disposition view)")
    field_kwargs = {"widget": Select2MultipleWidget}


@global_preferences_registry.register
class EnabledPlugins(MultipleChoicePreference):
    name = "enabled_plugins"
    verbose_name = _("Enabled plugins")
    default = [
        ephios.plugins.basesignup.apps.PluginApp.__module__,
        ephios.plugins.pages.apps.PluginApp.__module__,
    ]
    section = general_global_section
    required = False

    @staticmethod
    def get_choices():
        return [
            (plugin.module, mark_safe(f"<strong>{plugin.name}</strong>: {plugin.description}"))
            for plugin in plugins.get_all_plugins()
            if getattr(plugin, "visible", True)
        ]


@user_preferences_registry.register
class NotificationPreference(JSONPreference):
    name = "notifications"
    verbose_name = _("Notification preferences")
    section = notifications_user_section
    default = dict(
        zip(
            [not_type.slug for not_type in CORE_NOTIFICATION_TYPES],
            [[backend.slug for backend in CORE_NOTIFICATION_BACKENDS]]
            * len(CORE_NOTIFICATION_TYPES),
        )
    )


@event_type_preference_registry.register
class VisibleForPreference(ModelMultipleChoicePreference):
    name = "visible_for"
    verbose_name = _("Events of this type should by default be visible for")
    model = Group
    default = Group.objects.all()
    field_kwargs = {"widget": Select2MultipleWidget}


@event_type_preference_registry.register
class ResponsibleUsersPreference(ModelMultipleChoicePreference):
    name = "responsible_users"
    verbose_name = _("Users that are responsible for this event type by default")
    model = UserProfile
    default = UserProfile.objects.none()
    field_kwargs = {"widget": Select2MultipleWidget}


@event_type_preference_registry.register
class ResponsibleGroupsPreference(ModelMultipleChoicePreference):
    name = "responsible_groups"
    verbose_name = _("Groups that are responsible for this event type by default")
    model = Group
    default = Group.objects.none()
    field_kwargs = {"widget": Select2MultipleWidget}



# Generated by Django 3.1.2 on 2020-10-14 14:48

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0003_userprofile_calendar_token_squashed_0008_auto_20200925_1640"),
    ]

    operations = [
        migrations.AlterField(
            model_name="qualification",
            name="included_qualifications",
            field=models.ManyToManyField(
                blank=True, related_name="included_by", to="core.Qualification"
            ),
        ),
        migrations.AlterField(
            model_name="qualificationgrant",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="qualification_grants",
                to=settings.AUTH_USER_MODEL,
                verbose_name="user profile",
            ),
        ),
    ]



from django.contrib import messages
from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.timezone import get_default_timezone
from django.utils.translation import gettext as _
from django.views import View
from django.views.generic import DeleteView, TemplateView
from django.views.generic.detail import SingleObjectMixin

from ephios.core.forms.events import ShiftForm
from ephios.core.models import Event, Shift
from ephios.core.signup.methods import enabled_signup_methods, signup_method_from_slug
from ephios.extra.mixins import CustomPermissionRequiredMixin


class ShiftCreateView(CustomPermissionRequiredMixin, TemplateView):
    permission_required = "core.change_event"
    template_name = "core/shift_form.html"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.event = get_object_or_404(Event.all_objects, pk=self.kwargs.get("pk"))

    def get_permission_object(self):
        return self.event

    def get_shift_form(self):
        return ShiftForm(self.request.POST or None)

    def get_context_data(self, **kwargs):
        kwargs.setdefault("event", self.event)
        kwargs.setdefault("form", self.get_shift_form())
        kwargs.setdefault("configuration_form", "")
        return super().get_context_data(**kwargs)

    def post(self, *args, **kwargs):
        form = self.get_shift_form()

        try:
            signup_method = signup_method_from_slug(
                self.request.POST["signup_method_slug"], event=self.event
            )
        except KeyError:
            if not list(enabled_signup_methods()):
                form.add_error(
                    "signup_method_slug",
                    _("You must enable plugins providing signup methods to continue."),
                )
            return self.render_to_response(
                self.get_context_data(
                    form=form,
                )
            )
        except ValueError as e:
            raise ValidationError(e) from e
        else:
            configuration_form = signup_method.get_configuration_form(self.request.POST)
            if not all((form.is_valid(), configuration_form.is_valid())):
                return self.render_to_response(
                    self.get_context_data(
                        form=form,
                        configuration_form=signup_method.render_configuration_form(
                            form=configuration_form
                        ),
                    )
                )

            shift = form.save(commit=False)
            shift.event = self.event
            shift.signup_configuration = configuration_form.cleaned_data
            shift.save()
            if "addAnother" in self.request.POST:
                return redirect(
                    reverse("core:event_createshift", kwargs={"pk": self.kwargs.get("pk")})
                )
            try:
                self.event.activate()
                messages.success(
                    self.request,
                    _("The event {title} has been saved.").format(title=self.event.title),
                )
            except ValidationError as e:
                messages.error(self.request, e)
            return redirect(self.event.get_absolute_url())


class ShiftConfigurationFormView(CustomPermissionRequiredMixin, SingleObjectMixin, View):
    queryset = Event.all_objects
    permission_required = "core.change_event"
    pk_url_kwarg = "event_id"

    def get(self, request, *args, **kwargs):
        signup_method = signup_method_from_slug(self.kwargs.get("slug"), event=self.get_object())
        return HttpResponse(signup_method.render_configuration_form())


class ShiftUpdateView(CustomPermissionRequiredMixin, SingleObjectMixin, TemplateView):
    model = Shift
    template_name = "core/shift_form.html"
    permission_required = "core.change_event"

    def get_permission_object(self):
        return self.get_object().event

    def get_shift_form(self):
        return ShiftForm(
            self.request.POST or None,
            instance=self.object,
            initial={
                "date": self.object.meeting_time.date(),
                "meeting_time": self.object.meeting_time.astimezone(get_default_timezone()).time(),
                "start_time": self.object.start_time.astimezone(get_default_timezone()).time(),
                "end_time": self.object.end_time.astimezone(get_default_timezone()).time(),
            },
        )

    def get_configuration_form(self):
        return self.object.signup_method.render_configuration_form(data=self.request.POST or None)

    def get_context_data(self, **kwargs):
        self.object = self.get_object()
        kwargs.setdefault("event", self.object.event)
        kwargs.setdefault("form", self.get_shift_form())
        kwargs.setdefault("configuration_form", self.get_configuration_form())
        return super().get_context_data(**kwargs)

    def post(self, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_shift_form()

        try:
            signup_method = signup_method_from_slug(
                self.request.POST["signup_method_slug"], shift=self.object
            )
            configuration_form = signup_method.get_configuration_form(self.request.POST)
        except ValueError as e:
            raise ValidationError(e) from e
        if all([form.is_valid(), configuration_form.is_valid()]):
            shift = form.save(commit=False)
            shift.signup_configuration = configuration_form.cleaned_data
            shift.save()
            if "addAnother" in self.request.POST:
                return redirect(reverse("core:event_createshift", kwargs={"pk": shift.event.pk}))
            messages.success(
                self.request, _("The shift {shift} has been saved.").format(shift=shift)
            )
            return redirect(self.object.event.get_absolute_url())
        return self.render_to_response(
            self.get_context_data(
                form=form,
                configuration_form=signup_method.render_configuration_form(form=configuration_form),
            )
        )


class ShiftDeleteView(CustomPermissionRequiredMixin, DeleteView):
    permission_required = "core.change_event"
    model = Shift

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.object = self.get_object()

    def delete(self, request, *args, **kwargs):
        if self.object.event.shifts.count() == 1:
            messages.error(self.request, _("You cannot delete the last shift!"))
            return redirect(self.object.event.get_absolute_url())
        return super().delete(request, *args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, _("The shift has been deleted."))
        return self.object.event.get_absolute_url()

    def get_permission_object(self):
        return self.object.event



import dataclasses
import functools
import logging
from argparse import Namespace
from collections import OrderedDict
from typing import List, Optional

from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML, Field, Layout
from django import forms
from django.contrib import messages
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import transaction
from django.db.models import Q
from django.shortcuts import redirect
from django.template import Context, Template
from django.template.loader import get_template
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import date_format
from django.utils.functional import SimpleLazyObject, cached_property, classproperty
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView

from ephios.core.models import AbstractParticipation, Shift
from ephios.core.services.notifications.types import (
    ResponsibleConfirmedParticipationCustomizedNotification,
    ResponsibleConfirmedParticipationDeclinedNotification,
)
from ephios.core.signals import (
    check_participant_signup,
    participant_from_request,
    register_signup_methods,
)
from ephios.core.signup.participants import AbstractParticipant
from ephios.extra.utils import format_anything
from ephios.extra.widgets import CustomSplitDateTimeWidget

logger = logging.getLogger(__name__)


def installed_signup_methods():
    for _, methods in register_signup_methods.send_to_all_plugins(None):
        yield from methods


def enabled_signup_methods():
    for _, methods in register_signup_methods.send(None):
        yield from methods


def signup_method_from_slug(slug, shift=None, event=None):
    for method in installed_signup_methods():
        if method.slug == slug:
            return method(shift, event=event)
    raise ValueError(_("Signup Method '{slug}' was not found.").format(slug=slug))


class ParticipationError(ValidationError):
    """Superclass of errors used in the signup mechanism."""


class ActionDisallowedError(ValidationError):
    """Error to return if the participant cannot perform an action on a shift."""


class SignupDisallowedError(ActionDisallowedError):
    """
    Error to return if the participant cannot sign up for a shift,
    even if formally fit, because of certain other circumstances.
    """


class DeclineDisallowedError(ActionDisallowedError):
    """Error to return if the participant cannot decline a shift."""


class ParticipantUnfitError(ParticipationError):
    """
    Error to return if the participant is unfit for a shift,
    regardless of participation state or situation.
    """


def prevent_getting_participant_from_request_user(request):
    """
    To prevent plugin authors from accessing the participant using request.user, the SignupView
    uses this method to block access to `as_participant` of the request.user attribute value.
    """
    original_user = request.user

    class ProtectedUser(SimpleLazyObject):
        def as_participant(self):
            raise Exception("Access of request.user.as_participant in SignupViews is not allowed.")

    request.user = ProtectedUser(lambda: original_user)


def get_nonlocal_participant_from_request(request):
    for _, participant in participant_from_request.send(sender=None, request=request):
        if participant is not None:
            return participant
    raise PermissionDenied


class BaseParticipationForm(forms.ModelForm):
    individual_start_time = forms.SplitDateTimeField(
        label=_("Individual start time"), widget=CustomSplitDateTimeWidget, required=False
    )
    individual_end_time = forms.SplitDateTimeField(
        label=_("Individual end time"),
        widget=CustomSplitDateTimeWidget,
        required=False,
    )

    def clean_individual_start_time(self):
        if self.cleaned_data["individual_start_time"] == self.shift.start_time:
            return None
        return self.cleaned_data["individual_start_time"]

    def clean_individual_end_time(self):
        if self.cleaned_data["individual_end_time"] == self.shift.end_time:
            return None
        return self.cleaned_data["individual_end_time"]

    def clean(self):
        cleaned_data = super().clean()
        start = cleaned_data["individual_start_time"] or self.shift.start_time
        end = cleaned_data["individual_end_time"] or self.shift.end_time
        if end < start:
            self.add_error("individual_end_time", _("End time must not be before start time."))
        return cleaned_data

    class Meta:
        model = AbstractParticipation
        fields = ["individual_start_time", "individual_end_time", "comment"]

    def __init__(self, *args, **kwargs):
        instance = kwargs["instance"]
        self.shift = getattr(self, "shift", instance.shift)
        kwargs["initial"] = {
            **kwargs.get("initial", {}),
            "individual_start_time": instance.individual_start_time or self.shift.start_time,
            "individual_end_time": instance.individual_end_time or self.shift.end_time,
        }
        super().__init__(*args, **kwargs)

    def get_customization_notification_info(self):
        """
        Return a list of human readable messages for changed participation attributes responsibles should be informed about.
        This should not include the participation state, but customization aspects such as individual times, detailed disposition information, etc.
        """
        assert self.is_valid()
        info = []
        for time in ["start_time", "end_time"]:
            if (field_name := f"individual_{time}") in self.changed_data:
                info.append(
                    _("{label} was changed from {initial} to {current}.").format(
                        label=self.fields[field_name].label,
                        initial=date_format(
                            localtime(self.initial[field_name].astimezone()),
                            format="SHORT_DATETIME_FORMAT",
                        ),
                        current=date_format(
                            localtime(self.cleaned_data[field_name] or getattr(self.shift, time)),
                            format="TIME_FORMAT",
                        ),
                    )
                )

        return info


class BaseSignupForm(BaseParticipationForm):
    def _get_field_layout(self):
        return Layout(*(Field(name) for name in self.fields))

    def _get_buttons(self):
        if (
            p := self.participant.participation_for(self.method.shift)
        ) is not None and p.is_in_positive_state():
            buttons = [
                HTML(
                    f'<button class="btn btn-success mt-1 ms-1 float-end" type="submit" name="signup_choice" value="customize">{_("Save")}</button>'
                )
            ]
        else:
            buttons = [
                HTML(
                    f'<button class="btn btn-success mt-1 ms-1 float-end" type="submit" name="signup_choice" value="sign_up">{self.method.registration_button_text}</button>'
                )
            ]
        buttons.append(
            HTML(
                f'<a class="btn btn-secondary mt-1" href="{self.participant.reverse_event_detail(self.method.shift.event)}">{_("Cancel")}</a>'
            )
        )
        if self.method.can_decline(self.participant):
            buttons.append(
                HTML(
                    f'<button class="btn btn-secondary mt-1 ms-1 float-end" type="submit" name="signup_choice" value="decline">{_("Decline")}</button>'
                )
            )
        return buttons

    def __init__(self, *args, **kwargs):
        self.method = kwargs.pop("method")
        self.targets_positive_state = kwargs.pop("targets_positive_state")
        self.shift: Shift = self.method.shift
        self.participant: AbstractParticipant = kwargs.pop("participant")
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            self._get_field_layout(),
            FormActions(*self._get_buttons()),
        )

        if not self.shift.signup_method.configuration.user_can_customize_signup_times:
            self.fields["individual_start_time"].disabled = True
            self.fields["individual_end_time"].disabled = True

    def clean(self):
        cleaned_data = super().clean()
        self._validate_conflicting_participations(self, self.targets_positive_state)
        return cleaned_data

    def _validate_conflicting_participations(self, form, targets_positive_state):
        if not targets_positive_state:
            return
        if conflicts := get_conflicting_participations(
            participant=form.instance.participant,
            start_time=form.cleaned_data["individual_start_time"] or self.shift.start_time,
            end_time=form.cleaned_data["individual_end_time"] or self.shift.end_time,
            shift=self.shift,
            total=False,
        ):
            form.add_error("individual_start_time", "")
            form.add_error(
                "individual_end_time",
                _("You are already confirmed for other shifts at this time: {shifts}.").format(
                    shifts=", ".join(str(shift) for shift in conflicts)
                ),
            )


class BaseSignupView(FormView):
    """
    This View reacts to the signup or decline buttons being pressed using a POST request.
    Beware that request.user might be anonymous. You should only act on participants acquired
    using `get_participant`.
    """

    shift: Shift = ...
    method: "BaseSignupMethod" = ...
    form_class = BaseSignupForm
    template_name = "core/signup.html"

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(
            {
                "method": self.method,
                "participant": self.participant,
                "instance": self.participation,
                "targets_positive_state": self.request.POST.get("signup_choice") == "sign_up",
            }
        )
        return kwargs

    @cached_property
    def participation(self):
        return self.method.get_participation_for(self.participant)

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.participant: AbstractParticipant = kwargs["participant"]
        prevent_getting_participant_from_request_user(request)

    def form_valid(self, form):
        if (choice := self.request.POST.get("signup_choice")) is not None:
            if choice == "sign_up" and self.method.can_sign_up(self.participant):
                return self.signup_pressed(form)
            if choice == "customize" and self.method.can_customize_signup(self.participant):
                return self.customize_pressed(form)
            if choice == "decline":
                return self.decline_pressed(form)
            messages.error(self.request, _("This action is not allowed."))
            return redirect(self.participant.reverse_event_detail(self.shift.event))
        form.add_error(None, _("Form submission is missing the mode of signup."))
        return self.form_invalid(form)

    def customize_pressed(self, form):
        form.save()
        if claims := form.get_customization_notification_info():
            ResponsibleConfirmedParticipationCustomizedNotification(form.instance, claims).send()
        messages.success(self.request, _("Your participation was saved."))
        return redirect(self.participant.reverse_event_detail(self.shift.event))

    def signup_pressed(self, form):
        try:
            with transaction.atomic():
                participation = form.save()
                self.method.perform_signup(self.participant, participation, **form.cleaned_data)
        except ParticipationError as errors:
            for error in errors:
                messages.error(self.request, self.method.signup_error_message.format(error=error))
        else:
            messages.success(
                self.request,
                self.method.signup_success_message.format(shift=self.shift),
            )
        return redirect(self.participant.reverse_event_detail(self.shift.event))

    def decline_pressed(self, form):
        try:
            with transaction.atomic():
                participation = form.save()
                self.method.perform_decline(self.participant, participation, **form.cleaned_data)
        except ParticipationError as errors:
            for error in errors:
                messages.error(self.request, self.method.decline_error_message.format(error=error))
        else:
            messages.info(
                self.request, self.method.decline_success_message.format(shift=self.shift)
            )
        return redirect(self.participant.reverse_event_detail(self.shift.event))


def check_event_is_active(method, participant):
    if not method.shift.event.active:
        return ActionDisallowedError(_("The event is not active."))


def check_participation_state_for_signup(method, participant):
    participation = participant.participation_for(method.shift)
    if participation is not None:
        if participation.state == AbstractParticipation.States.RESPONSIBLE_REJECTED:
            return SignupDisallowedError(_("You have been rejected."))


def check_participation_state_for_decline(method, participant):
    participation = participant.participation_for(method.shift)
    if participation is not None:
        if (
            participation.state == AbstractParticipation.States.CONFIRMED
            and not method.configuration.user_can_decline_confirmed
        ):
            return DeclineDisallowedError(_("You cannot decline by yourself."))
        if participation.state == AbstractParticipation.States.RESPONSIBLE_REJECTED:
            return DeclineDisallowedError(_("You have been rejected."))
        if participation.state == AbstractParticipation.States.USER_DECLINED:
            return DeclineDisallowedError(_("You have already declined participating."))


def check_inside_signup_timeframe(method, participant):
    last_time = method.shift.end_time
    if method.configuration.signup_until is not None:
        last_time = min(last_time, method.configuration.signup_until)
    if timezone.now() > last_time:
        return ActionDisallowedError(_("The signup period is over."))


def check_participant_age(method, participant):
    minimum_age = method.configuration.minimum_age
    day = method.shift.start_time.date()
    age = participant.get_age(day)
    if minimum_age is not None and age is not None and age < minimum_age:
        return ParticipantUnfitError(
            _("You are too young. The minimum age is {age}.").format(age=minimum_age)
        )


def check_participant_signup_signal(method, participant):
    errors = []
    for _, result in check_participant_signup.send(None, method=method, participant=participant):
        if result is not None:
            errors.append(result)
    return errors


def get_conflicting_participations(
    participant: AbstractParticipant, shift: Shift, start_time=None, end_time=None, total=False
):
    """
    Return a queryset of participations of a participant in conflict with
    a (potential) participation specified by the arguments.

        Parameters:
            participant: AbstractParticipant to check conflict for
            shift: conflicting participations for this shift
            start_time, end_time: specify times other than the default times of `shift`
            total (bool): If True, only return conflicts that can't be
                          resolved by only participating in only part of `shift`.

        Returns:
            a queryset of participations
    """
    start_time = start_time or shift.start_time
    end_time = end_time or shift.end_time
    qs = participant.all_participations().filter(
        ~Q(shift=shift)
        & Q(state=AbstractParticipation.States.CONFIRMED)
        & Q(start_time__lt=end_time, end_time__gt=start_time)
    )
    if total:
        qs = qs.filter(start_time__lte=shift.start_time, end_time__gte=shift.end_time)
    return qs


def check_conflicting_participations(method, participant):
    start_time, end_time = method.shift.start_time, method.shift.end_time
    if participation := participant.participation_for(method.shift):
        start_time, end_time = participation.start_time, participation.end_time

    # if users can provide individual times, only total conflicts should block signup
    total = getattr(method.configuration, "user_can_customize_signup_times", False)
    if conflicts := get_conflicting_participations(
        participant=participant,
        shift=method.shift,
        start_time=start_time,
        end_time=end_time,
        total=total,
    ):
        return SignupDisallowedError(
            _("You are already confirmed for other shifts at this time: {shifts}.").format(
                shifts=", ".join(str(shift) for shift in conflicts)
            )
        )


class BaseSignupMethodConfigurationForm(forms.Form):
    minimum_age = forms.IntegerField(
        required=False, min_value=1, max_value=999, initial=None, label=_("Minimum age")
    )
    signup_until = forms.SplitDateTimeField(
        required=False,
        widget=CustomSplitDateTimeWidget,
        initial=None,
        label=_("Signup until"),
    )
    user_can_decline_confirmed = forms.BooleanField(
        label=_("Confirmed users can decline by themselves"),
        required=False,
        help_text=_("only if the signup timeframe has not ended"),
    )
    user_can_customize_signup_times = forms.BooleanField(
        label=_("Users can provide individual start and end times"),
        required=False,
        initial=True,
    )


class BaseSignupMethod:
    # pylint: disable=too-many-public-methods

    @property
    def slug(self):
        raise NotImplementedError()

    @property
    def verbose_name(self):
        raise NotImplementedError()

    @property
    def disposition_participation_form_class(self):
        """
        This form will be used for participations in disposition.
        Set to None if you don't want to support the default disposition.
        """
        from .disposition import BaseDispositionParticipationForm

        return BaseDispositionParticipationForm

    configuration_form_class = BaseSignupMethodConfigurationForm

    @property
    def signup_view_class(self):
        return BaseSignupView

    @property
    def shift_state_template_name(self):
        raise NotImplementedError

    description = """"""

    # use _ == gettext_lazy!
    registration_button_text = _("Sign up")
    signup_success_message = _("You have successfully signed up for {shift}.")
    signup_error_message = _("Signing up failed: {error}")
    decline_success_message = _("You have successfully declined {shift}.")
    decline_error_message = _("Declining failed: {error}")

    uses_requested_state = True

    def __init__(self, shift, event=None):
        self.shift = shift
        self.event = getattr(shift, "event", event)
        self.configuration = Namespace(
            **{
                name: field.initial
                for name, field in self.configuration_form_class.base_fields.items()
            }
        )
        if shift is not None:
            for key, value in shift.signup_configuration.items():
                setattr(self.configuration, key, value)

    @cached_property
    def signup_view(self):
        return self.signup_view_class.as_view(method=self, shift=self.shift)

    @property
    def _signup_checkers(self):
        """Return a list of methods that check if the participant can sign up for the shift."""
        return [
            check_event_is_active,
            check_participation_state_for_signup,
            check_inside_signup_timeframe,
            check_conflicting_participations,
            check_participant_age,
            check_participant_signup_signal,
        ]

    @property
    def _decline_checkers(self):
        """Return a list of methods that check if the participant can decline the shift."""
        return [
            check_event_is_active,
            check_participation_state_for_decline,
            check_inside_signup_timeframe,
        ]

    def _run_checkers(self, participant, checkers) -> List[ParticipationError]:
        errors = []
        for checker in checkers:
            # checkers can return None, a single ParticipationError, or a list of them
            if (error := checker(self, participant)) is not None:
                if isinstance(error, list):
                    errors.extend(error)
                else:
                    errors.append(error)
        return errors

    @functools.lru_cache()
    def get_signup_errors(self, participant) -> List[ParticipationError]:
        """Return a list of ParticipationErrors that describe reasons for not being able to sign up."""
        return self._run_checkers(participant, self._signup_checkers)

    @functools.lru_cache()
    def get_decline_errors(self, participant):
        """Return a list of ParticipationErrors that describe reasons for not being able to decline."""
        return self._run_checkers(participant, self._decline_checkers)

    @functools.lru_cache()
    def get_participant_errors(self, participant) -> List[ParticipationError]:
        """
        Return errors for whether the participant fulfills the requirements set by the signup methods for participation.
        This runs a subset of the check for `get_signup_errors`.
        """
        return list(
            filter(
                lambda error: isinstance(error, ParticipantUnfitError),
                self.get_signup_errors(participant),
            )
        )

    def can_decline(self, participant):
        """
        Return whether the participant is allowed to decline.
        """
        return not self.get_decline_errors(participant)

    def can_sign_up(self, participant):
        """
        Return whether the participant is allowed to perform signup.
        """
        signupable_state = (
            p := participant.participation_for(self.shift)
        ) is None or not p.is_in_positive_state()
        return signupable_state and not self.get_signup_errors(participant)

    def can_customize_signup(self, participant):
        """
        Return whether the participant gets shown the option to customize their participation.
        """
        positive_state = (
            p := participant.participation_for(self.shift)
        ) is not None and p.is_in_positive_state()

        if positive_state:
            # If in positive state, check that you can decline and then sign up again.
            return self.can_decline(participant) and not self.get_signup_errors(participant)
        return not self.get_signup_errors(participant)

    def has_customized_signup(self, participation):
        """
        Return whether the participation was customized in a way specific to this signup method.
        """
        # This method should most likely check the participation's data attribute for modifications it has done.
        # 'Customized' in this context means that the dispositioning person should give special attention to this participation.
        return False

    def get_participation_for(self, participant) -> AbstractParticipation:
        return participant.participation_for(self.shift) or participant.new_participation(
            self.shift
        )

    def perform_signup(
        self, participant: AbstractParticipant, participation=None, **kwargs
    ) -> AbstractParticipation:
        """
        Creates and/or configures a participation object for a given participant and sends out notifications.
        Passes the participation and kwargs to configure_participation to do configuration specific to the signup method.
        """
        from ephios.core.services.notifications.types import (
            ResponsibleParticipationRequestedNotification,
        )

        if errors := self.get_signup_errors(participant):
            raise ParticipationError(errors)
        participation = participation or self.get_participation_for(participant)
        participation = self._configure_participation(participation, **kwargs)
        participation.save()
        ResponsibleParticipationRequestedNotification(participation).send()
        return participation

    def perform_decline(self, participant, participation=None, **kwargs):
        """Create and configure a declining participation object for the given participant. `kwargs` may contain further instructions from a e.g. a form."""
        if errors := self.get_decline_errors(participant):
            raise ParticipationError(errors)
        participation = participation or self.get_participation_for(participant)
        participation.state = AbstractParticipation.States.USER_DECLINED
        participation.save()
        ResponsibleConfirmedParticipationDeclinedNotification(participation).send()
        return participation

    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        """
        Configure the given participation object for signup according to the method's configuration.
        You need at least to set the participations state, as that is not done with the participation form.
        `kwargs` contains the signup form's cleaned_data.
        """
        raise NotImplementedError

    def get_signup_info(self):
        """
        Return key/value pairs about the configuration to show in exports etc.
        """
        form_class = self.configuration_form_class
        return OrderedDict(
            {
                label: getattr(form_class, f"format_{name}", format_anything)(value)
                for name, field in form_class.base_fields.items()
                if (label := field.label) and (value := getattr(self.configuration, name))
            }
        )

    def get_participant_count_bounds(self):
        """
        Return a typle of min, max for how many participants are allowed for the shift.
        Use None for any value if it is not specifiable.
        The default implementation returns None, None to signify both values are not specified.
        """
        return None, None

    def get_signup_stats(self) -> "SignupStats":
        """
        Return an instance of SignupStats for the shift.
        """
        min_count, max_count = self.get_participant_count_bounds()
        participations = list(self.shift.participations.all())
        confirmed_count = sum(
            p.state == AbstractParticipation.States.CONFIRMED for p in participations
        )
        return SignupStats(
            requested_count=sum(
                p.state == AbstractParticipation.States.REQUESTED for p in participations
            ),
            confirmed_count=confirmed_count,
            missing=max(min_count - confirmed_count, 0) if min_count else 0,
            free=max(max_count - confirmed_count, 0) if max_count else None,
            min_count=min_count,
            max_count=max_count,
        )

    def render(self, context):
        """
        Render the state/participations of the shift.
        Match the signature of template.render for use with the include template tag:
        {% include shift.signup_method %}
        By default, this loads `shift_state_template_name` and renders it using context from `get_shift_state_context_data`.
        """
        with context.update(self.get_shift_state_context_data(context.request)):
            return get_template(self.shift_state_template_name).template.render(context)

    def get_shift_state_context_data(self, request, **kwargs):
        """
        Additionally to the context of the event detail view, provide context for rendering `shift_state_template_name`.
        """
        kwargs["shift"] = self.shift
        kwargs["participations"] = self.shift.participations.filter(
            state__in={
                AbstractParticipation.States.REQUESTED,
                AbstractParticipation.States.CONFIRMED,
            }
        ).order_by("-state")
        if self.disposition_participation_form_class is not None:
            kwargs["disposition_url"] = (
                reverse("core:shift_disposition", kwargs=dict(pk=self.shift.pk))
                if request.user.has_perm("core.change_event", obj=self.shift.event)
                else None
            )
        return kwargs

    def render_shift_state(self, context):
        """
        Render html that will be shown in the shift info box.
        Use it to inform about the current state of the shift and participations.
        """
        return ""

    def get_participation_display(self):
        """
        Returns a displayable representation of participation that can be rendered into a table (e.g. for pdf export).
        Must return a list of participations or empty slots. Each element of the list has to be a list of a fixed
        size where each entry is rendered to a separate column.
        Ex.: [["participant1_name", "participant1_qualification"], ["participant2_name", "participant2_qualification"]]
        """
        return [
            [f"{participant.first_name} {participant.last_name}"]
            for participant in self.shift.get_participants()
        ]

    def get_configuration_form(self, *args, **kwargs):
        if self.shift is not None:
            kwargs.setdefault("initial", self.configuration.__dict__)
        form = self.configuration_form_class(*args, **kwargs)
        return form

    def render_configuration_form(self, *args, form=None, **kwargs):
        form = form or self.get_configuration_form(*args, **kwargs)
        template = Template(
            template_string="{% load crispy_forms_filters %}{{ form|crispy }}"
        ).render(Context({"form": form}))
        return template


@dataclasses.dataclass()
class SignupStats:
    requested_count: int
    confirmed_count: int
    missing: int
    free: Optional[int]  # None means infinite free
    min_count: Optional[int]  # None means no min specified
    max_count: Optional[int]  # None means infinite max

    @classproperty
    def ZERO(cls):  # pylint: disable=no-self-argument
        return SignupStats(
            requested_count=0,
            confirmed_count=0,
            missing=0,
            free=0,
            min_count=None,
            max_count=0,
        )

    def __add__(self, other: "SignupStats"):
        free = self.free + other.free if self.free is not None and other.free is not None else None
        missing = self.missing + other.missing
        min_count = (
            (self.min_count or 0) + (other.min_count or 0)
            if self.min_count is not None or other.min_count is not None
            else None
        )
        max_count = (
            self.max_count + other.max_count
            if self.max_count is not None and other.max_count is not None
            else None
        )
        return SignupStats(
            requested_count=self.requested_count + other.requested_count,
            confirmed_count=self.confirmed_count + other.confirmed_count,
            missing=missing,
            free=free,
            min_count=min_count,
            max_count=max_count,
        )



from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext as _

from ephios.core.signals import administration_settings_section


@receiver(
    administration_settings_section,
    dispatch_uid="ephios.plugins.qualification_management.signals.qualifications_settings_section",
)
def qualifications_settings_section(sender, request, **kwargs):
    return [
        {
            "label": _("Qualifications"),
            "url": reverse("qualification_management:settings_qualification_list"),
            "active": request.resolver_match.url_name.startswith("settings_qualification"),
        },
    ]



from django.urls import path

from ephios.plugins.qualification_management.views import (
    QualificationCategorySetUpdateView,
    QualificationCreateView,
    QualificationDeleteView,
    QualificationExportFixtureView,
    QualificationImportView,
    QualificationListView,
    QualificationReassignmentView,
    QualificationUpdateView,
)

app_name = "qualification_management"
urlpatterns = [
    path(
        "settings/qualifications/",
        QualificationListView.as_view(),
        name="settings_qualification_list",
    ),
    path(
        "settings/qualifications/import/",
        QualificationImportView.as_view(),
        name="settings_qualification_import",
    ),
    path(
        "settings/qualifications/<int:pk>/edit/",
        QualificationUpdateView.as_view(),
        name="settings_qualification_edit",
    ),
    path(
        "settings/qualifications/<int:pk>/delete/",
        QualificationDeleteView.as_view(),
        name="settings_qualification_delete",
    ),
    path(
        "settings/qualifications/create/",
        QualificationCreateView.as_view(),
        name="settings_qualification_create",
    ),
    path(
        "settings/qualifications/categories/",
        QualificationCategorySetUpdateView.as_view(),
        name="settings_qualification_categories",
    ),
    path(
        "settings/qualifications/reassign/",
        QualificationReassignmentView.as_view(),
        name="settings_qualification_reassignment",
    ),
    path(
        "settings/qualifications/export/",
        QualificationExportFixtureView.as_view(),
        name="settings_qualification_export",
    ),
]



import datetime
import functools
import secrets
import uuid
from datetime import date
from itertools import chain

import guardian.mixins
from django.contrib.auth import get_user_model
from django.contrib.auth.base_user import AbstractBaseUser, BaseUserManager
from django.contrib.auth.models import Group, PermissionsMixin
from django.db import models, transaction
from django.db.models import (
    BooleanField,
    CharField,
    DateField,
    EmailField,
    ExpressionWrapper,
    F,
    ForeignKey,
    Max,
    Model,
    Q,
    Sum,
)
from django.db.models.functions import TruncDate
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ephios.extra.fields import EndOfDayDateTimeField
from ephios.extra.json import CustomJSONDecoder, CustomJSONEncoder
from ephios.extra.widgets import CustomDateInput
from ephios.modellogging.log import (
    ModelFieldsLogConfig,
    add_log_recorder,
    register_model_for_logging,
)
from ephios.modellogging.recorders import FixedMessageLogRecorder, M2MLogRecorder


class UserProfileManager(BaseUserManager):
    def create_user(
        self,
        email,
        first_name,
        last_name,
        date_of_birth,
        password=None,
    ):
        # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
        user = self.model(
            email=email,
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
        )
        user.set_password(password)
        user.save()
        return user

    def create_superuser(
        self,
        email,
        first_name,
        last_name,
        date_of_birth,
        password=None,
    ):
        user = self.create_user(
            email=email,
            password=password,
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
        )
        user.is_superuser = True
        user.is_staff = True
        user.save()
        return user


class VisibleUserProfileManager(BaseUserManager):
    def get_queryset(self):
        return super().get_queryset().filter(is_visible=True)


class UserProfile(guardian.mixins.GuardianUserMixin, PermissionsMixin, AbstractBaseUser):
    email = EmailField(_("email address"), unique=True)
    is_active = BooleanField(default=True, verbose_name=_("Active"))
    is_visible = BooleanField(default=True, verbose_name=_("Visible"))
    is_staff = BooleanField(default=False, verbose_name=_("Staff user"))
    first_name = CharField(_("first name"), max_length=254)
    last_name = CharField(_("last name"), max_length=254)
    date_of_birth = DateField(_("date of birth"))
    phone = CharField(_("phone number"), max_length=254, blank=True)
    calendar_token = CharField(_("calendar token"), max_length=254, default=secrets.token_urlsafe)

    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = [
        "first_name",
        "last_name",
        "date_of_birth",
    ]

    objects = VisibleUserProfileManager()
    all_objects = UserProfileManager()

    class Meta:
        verbose_name = _("user profile")
        verbose_name_plural = _("user profiles")
        db_table = "userprofile"
        base_manager_name = "all_objects"
        default_manager_name = "all_objects"

    def get_full_name(self):
        return self.first_name + " " + self.last_name

    def __str__(self):
        return self.get_full_name()

    def get_short_name(self):
        return self.first_name

    @property
    def age(self):
        today, born = date.today(), self.date_of_birth
        return today.year - born.year - ((today.month, today.day) < (born.month, born.day))

    @property
    def is_minor(self):
        return self.age < 18

    def as_participant(self):
        from ephios.core.signup.participants import LocalUserParticipant

        return LocalUserParticipant(
            first_name=self.first_name,
            last_name=self.last_name,
            qualifications=self.qualifications,
            date_of_birth=self.date_of_birth,
            email=self.email if self.is_active else None,
            user=self,
        )

    @property
    def qualifications(self):
        return Qualification.objects.filter(
            pk__in=self.qualification_grants.unexpired().values_list("qualification_id", flat=True)
        ).annotate(
            expires=Max(F("grants__expires"), filter=Q(grants__user=self)),
        )

    def get_workhour_items(self):
        from ephios.core.models import AbstractParticipation

        participations = (
            self.participations.filter(state=AbstractParticipation.States.CONFIRMED)
            .annotate(
                duration=ExpressionWrapper(
                    (F("end_time") - F("start_time")),
                    output_field=models.DurationField(),
                ),
                date=ExpressionWrapper(TruncDate(F("start_time")), output_field=DateField()),
                reason=F("shift__event__title"),
            )
            .values("duration", "date", "reason")
        )
        workinghours = self.workinghours_set.annotate(duration=F("hours")).values(
            "duration", "date", "reason"
        )
        hour_sum = (
            participations.aggregate(Sum("duration"))["duration__sum"] or datetime.timedelta()
        ) + datetime.timedelta(
            hours=float(workinghours.aggregate(Sum("duration"))["duration__sum"] or 0)
        )
        return hour_sum, list(sorted(chain(participations, workinghours), key=lambda k: k["date"]))


register_model_for_logging(
    UserProfile,
    ModelFieldsLogConfig(
        unlogged_fields={"id", "password", "calendar_token", "last_login"},
    ),
)

register_model_for_logging(
    Group,
    ModelFieldsLogConfig(
        unlogged_fields={"id", "permissions"},
        initial_recorders_func=lambda group: [
            M2MLogRecorder(UserProfile.groups.field, reverse=True, verbose_name=_("Users")),
        ],
    ),
)


class QualificationCategoryManager(models.Manager):
    def get_by_natural_key(self, category_uuid, *args):
        return self.get(uuid=category_uuid)


class QualificationCategory(Model):
    uuid = models.UUIDField("UUID", unique=True, default=uuid.uuid4)
    title = CharField(_("title"), max_length=254)

    objects = QualificationCategoryManager()

    class Meta:
        verbose_name = _("qualification track")
        verbose_name_plural = _("qualification tracks")
        db_table = "qualificationcategory"

    def __str__(self):
        return str(self.title)

    def natural_key(self):
        return (self.uuid, self.title)


class QualificationManager(models.Manager):
    def get_by_natural_key(self, qualification_uuid, *args):
        return self.get(uuid=qualification_uuid)


class Qualification(Model):
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, verbose_name="UUID")
    title = CharField(_("title"), max_length=254)
    abbreviation = CharField(max_length=254)
    category = ForeignKey(
        QualificationCategory,
        on_delete=models.CASCADE,
        related_name="qualifications",
        verbose_name=_("category"),
    )
    includes = models.ManyToManyField(
        "self", related_name="included_by", symmetrical=False, blank=True
    )
    is_imported = models.BooleanField(verbose_name=_("imported"), default=True)

    objects = QualificationManager()

    def __eq__(self, other):
        return self.uuid == other.uuid if other else False

    def __hash__(self):
        return hash(self.uuid)

    class Meta:
        verbose_name = _("qualification")
        verbose_name_plural = _("qualifications")
        db_table = "qualification"

    def __str__(self):
        return str(self.title)

    def natural_key(self):
        return (self.uuid, self.title)

    natural_key.dependencies = ["core.QualificationCategory"]

    @classmethod
    def collect_all_included_qualifications(cls, given_qualifications) -> set:
        """We collect using breadth first search with one query for every layer of inclusion."""
        all_qualifications = set(given_qualifications)
        current = set(given_qualifications)
        while current:
            new = (
                Qualification.objects.filter(included_by__in=current)
                .exclude(id__in=(q.id for q in all_qualifications))
                .distinct()
            )
            all_qualifications |= set(new)
            current = new
        return all_qualifications


class CustomQualificationGrantQuerySet(models.QuerySet):
    # Available on both Manager and QuerySet.
    def unexpired(self):
        return self.exclude(expires__isnull=False, expires__lt=timezone.now())


class ExpirationDateField(models.DateTimeField):
    """
    A model datetime field whose formfield is an EndOfDayDateTimeField
    """

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "widget": CustomDateInput,
                "form_class": EndOfDayDateTimeField,
                **kwargs,
            }
        )


class QualificationGrant(Model):
    qualification = ForeignKey(
        Qualification,
        on_delete=models.CASCADE,
        verbose_name=_("qualification"),
        related_name="grants",
    )
    user = ForeignKey(
        get_user_model(),
        related_name="qualification_grants",
        on_delete=models.CASCADE,
        verbose_name=_("user profile"),
    )
    expires = ExpirationDateField(_("expiration date"), blank=True, null=True)

    objects = CustomQualificationGrantQuerySet.as_manager()

    def __str__(self):
        return f"{self.qualification!s} {_('for')} {self.user!s}"

    class Meta:
        unique_together = [["qualification", "user"]]  # issue #218
        db_table = "qualificationgrant"
        verbose_name = _("Qualification grant")


register_model_for_logging(
    QualificationGrant,
    ModelFieldsLogConfig(attach_to_func=lambda grant: (UserProfile, grant.user_id)),
)


class Consequence(Model):
    slug = models.CharField(max_length=255)
    data = models.JSONField(default=dict, encoder=CustomJSONEncoder, decoder=CustomJSONDecoder)

    user = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        verbose_name=_("affected user"),
        null=True,
        related_name="affecting_consequences",
    )

    class States(models.TextChoices):
        NEEDS_CONFIRMATION = "needs_confirmation", _("needs confirmation")
        EXECUTED = "executed", _("executed")
        FAILED = "failed", _("failed")
        DENIED = "denied", _("denied")

    state = models.TextField(
        max_length=31,
        choices=States.choices,
        default=States.NEEDS_CONFIRMATION,
        verbose_name=_("State"),
    )

    class Meta:
        db_table = "consequence"
        verbose_name = _("Consequence")

    @property
    def handler(self):
        from ephios.core import consequences

        return consequences.consequence_handler_from_slug(self.slug)

    def confirm(self, user):
        from ephios.core.consequences import ConsequenceError

        if self.state not in {
            self.States.NEEDS_CONFIRMATION,
            self.States.DENIED,
            self.States.FAILED,
        }:
            raise ConsequenceError(_("Consequence was executed already."))

        try:
            with transaction.atomic():
                self.handler.execute(self)
                from ephios.core.services.notifications.types import ConsequenceApprovedNotification

                if user != self.user:
                    ConsequenceApprovedNotification.send(self)
        except Exception as e:  # pylint: disable=broad-except
            self.state = self.States.FAILED
            add_log_recorder(
                self,
                FixedMessageLogRecorder(
                    label=_("Reason"),
                    message=str(e),
                ),
            )
            raise ConsequenceError(str(e)) from e
        else:
            self.state = self.States.EXECUTED
        finally:
            self.save()

    def deny(self, user):
        from ephios.core.consequences import ConsequenceError

        if self.state not in {self.States.NEEDS_CONFIRMATION, self.States.FAILED}:
            raise ConsequenceError(_("Consequence was executed or denied already."))
        self.state = self.States.DENIED
        self.save()
        from ephios.core.services.notifications.types import ConsequenceDeniedNotification

        if user != self.user:
            ConsequenceDeniedNotification.send(self)

    def render(self):
        return self.handler.render(self)

    def __str__(self):
        return self.render()

    def attach_log_to_object(self):
        if self.user_id:
            return UserProfile, self.user_id
        return Consequence, self.id


register_model_for_logging(
    Consequence,
    ModelFieldsLogConfig(
        unlogged_fields=["id", "slug", "user", "data"],
        attach_to_func=lambda consequence: consequence.attach_log_to_object(),
    ),
)


class WorkingHours(Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    hours = models.DecimalField(decimal_places=2, max_digits=7)
    reason = models.CharField(max_length=1024, default="")
    date = models.DateField()

    class Meta:
        db_table = "workinghours"


class Notification(Model):
    slug = models.SlugField(max_length=255)
    user = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        verbose_name=_("affected user"),
        null=True,
    )
    failed = models.BooleanField(default=False)
    data = models.JSONField(
        blank=True, default=dict, encoder=CustomJSONEncoder, decoder=CustomJSONDecoder
    )
    created_at = models.DateTimeField(auto_now_add=True)

    @functools.cached_property
    def notification_type(self):
        from ephios.core.services.notifications.types import notification_type_from_slug

        return notification_type_from_slug(self.slug)

    @property
    def subject(self):
        return self.notification_type.get_subject(self)

    def as_plaintext(self):
        return self.notification_type.as_plaintext(self)

    def as_html(self):
        return self.notification_type.as_html(self)

    def get_url(self):
        return self.notification_type.get_url(self)



import logging

from django.conf import settings
from django.utils import timezone

from ephios.core.models import AbstractParticipation
from ephios.core.services.password_reset import logger

logger = logging.getLogger(__name__)


def send_participation_finished(sender, **kwargs):
    """
    This method is registered in signals.py as a receiver of ``periodic_signal``.
    """
    from ephios.core.signals import participation_finished

    relevant_participations = AbstractParticipation.objects.filter(
        state=AbstractParticipation.States.CONFIRMED,
        shift__end_time__lt=timezone.now(),
        finished=False,
    ).select_related("shift", "shift__event")
    for participation in relevant_participations:
        participation.finished = True
        participation.save()
        for __, result in participation_finished.send_robust(None, participation=participation):
            if isinstance(result, Exception):
                if settings.DEBUG:
                    raise result
                logger.exception(
                    "Error while dispatching participation_finished: %s", exc_info=result
                )



import itertools

from django import template
from django.db import models
from django.utils.safestring import mark_safe

from ephios.modellogging.log import LOGGED_MODELS
from ephios.modellogging.models import LogEntry

register = template.Library()


@register.filter(name="linkify_absolute_url")
def linkify_absolute_url(instance):
    if hasattr(instance, "get_absolute_url"):
        return mark_safe(
            f'<a class="log-absolute-url" href="{instance.get_absolute_url()}">{str(instance)}</a>'
        )
    return instance


@register.filter(name="related_logentries")
def related_logentries(instance):
    return LOGGED_MODELS[type(instance)].related_logentries(instance)


@register.filter(name="visible_logentries")
def visible_logentries(user):
    if user.has_perm("modellogging.view_logentry"):
        return LogEntry.objects.all()[:5]
    else:
        return LogEntry.objects.none()


@register.filter(name="group_logentries")
def group_logentries(logentries):
    if isinstance(logentries, models.QuerySet):
        logentries.select_related("user")
    yield from (
        list(group)
        for key, group in itertools.groupby(
            logentries,
            lambda entry: entry.request_id or entry.pk,
        )
    )



from django.dispatch import receiver

from ephios.core.signals import event_forms, participant_from_request
from ephios.plugins.guests.forms import EventAllowGuestsForm
from ephios.plugins.guests.models import GuestUser


@receiver(
    participant_from_request,
    dispatch_uid="ephios.plugins.guests.signals.guest_participant_from_request",
)
def guest_participant_from_request(sender, request, **kwargs):
    if token := request.resolver_match.kwargs.get("guest_access_token"):
        try:
            return GuestUser.objects.get(access_token=token).as_participant()
        except GuestUser.DoesNotExist:
            pass


@receiver(
    event_forms,
    dispatch_uid="ephios.plugins.guests.signals.guests_event_forms",
)
def guests_event_forms(sender, event, request, **kwargs):
    return [EventAllowGuestsForm(request.POST or None, event=event, request=request)]



# Generated by Django 3.1.7 on 2021-04-03 21:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("guests", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="guestparticipation",
            name="guest_user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="guests.guestuser",
                verbose_name="guest participant",
            ),
        ),
    ]



from django.urls import path

from ephios.plugins.pages.views import (
    PageCreateView,
    PageDeleteView,
    PageListView,
    PageUpdateView,
    PageView,
)

app_name = "pages"
urlpatterns = [
    path("page/<slug:slug>/", PageView.as_view(), name="page_detail"),
    path("settings/pages/", PageListView.as_view(), name="settings_page_list"),
    path("settings/pages/create/", PageCreateView.as_view(), name="settings_page_create"),
    path("settings/pages/<int:pk>/edit/", PageUpdateView.as_view(), name="settings_page_edit"),
    path("settings/pages/<int:pk>/delete/", PageDeleteView.as_view(), name="settings_page_delete"),
]



# Generated by Django 3.1.7 on 2021-03-09 15:35

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0004_auto_20210305_1855"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="event",
            name="mail_updates",
        ),
    ]



# Generated by Django 3.2.3 on 2021-05-28 20:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0008_auto_20210403_2348"),
    ]

    operations = [
        migrations.AlterField(
            model_name="abstractparticipation",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="consequence",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="event",
            name="active",
            field=models.BooleanField(default=False, verbose_name="active"),
        ),
        migrations.AlterField(
            model_name="event",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="eventtype",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="eventtypepreference",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="qualificationcategory",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="qualificationgrant",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="shift",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="workinghours",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-03 10:58

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0010_placeholderparticipation"),
    ]

    operations = [
        migrations.AddField(
            model_name="qualification",
            name="is_imported",
            field=models.BooleanField(default=True, verbose_name="imported"),
        ),
        migrations.AlterField(
            model_name="qualification",
            name="uuid",
            field=models.UUIDField(default=uuid.uuid4, unique=True, verbose_name="UUID"),
        ),
        migrations.AlterField(
            model_name="qualificationcategory",
            name="uuid",
            field=models.UUIDField(default=uuid.uuid4, unique=True, verbose_name="UUID"),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from ephios.core.models import Event, Qualification
from ephios.core.models.users import ExpirationDateField
from ephios.modellogging.log import ModelFieldsLogConfig, register_model_for_logging


class EventAutoQualificationConfiguration(models.Model):
    event = models.OneToOneField(
        Event, on_delete=models.CASCADE, related_name="auto_qualification_config"
    )

    qualification = models.ForeignKey(
        Qualification, on_delete=models.CASCADE, verbose_name=_("Qualification")
    )
    expiration_date = ExpirationDateField(verbose_name=_("Expiration date"), null=True, blank=True)

    class Modes(models.IntegerChoices):
        ANY_SHIFT = 1, _("any shift")
        EVERY_SHIFT = 2, _("every shift")
        LAST_SHIFT = 3, _("last shift")

    mode = models.IntegerField(
        choices=Modes.choices,
        default=Modes.ANY_SHIFT,
        verbose_name=_("Required attendance"),
    )

    extend_only = models.BooleanField(
        default=False, verbose_name=_("Only extend or reactivate existing qualification")
    )
    needs_confirmation = models.BooleanField(
        default=True, verbose_name=_("Qualification must be confirmed afterwards")
    )

    def __str__(self):
        return str(self.qualification or self.event)

    class Meta:
        verbose_name = _("event auto qualification configuration")


register_model_for_logging(
    EventAutoQualificationConfiguration,
    ModelFieldsLogConfig(
        unlogged_fields=["id", "event"],
        attach_to_func=lambda instance: (Event, instance.event_id),
    ),
)



# Generated by Django 3.1.6 on 2021-02-07 12:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0007_auto_20210110_0050"),
    ]

    operations = [
        migrations.AlterModelTable(
            name="consequence",
            table="consequence",
        ),
        migrations.AlterModelTable(
            name="qualification",
            table="qualification",
        ),
        migrations.AlterModelTable(
            name="qualificationcategory",
            table="qualificationcategory",
        ),
        migrations.AlterModelTable(
            name="qualificationgrant",
            table="qualificationgrant",
        ),
        migrations.AlterModelTable(
            name="userprofile",
            table="userprofile",
        ),
        migrations.AlterModelTable(
            name="workinghours",
            table="workinghours",
        ),
    ]



# Generated by Django 3.2.8 on 2021-10-27 20:03

import django.db.models.manager
from django.db import migrations, models

import ephios.core.models.users


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0012_rename_included_qualifications_qualification_includes"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="event",
            options={
                "base_manager_name": "all_objects",
                "default_manager_name": "all_objects",
                "permissions": [("view_past_event", "Can view past events")],
                "verbose_name": "event",
                "verbose_name_plural": "events",
            },
        ),
        migrations.AlterModelOptions(
            name="userprofile",
            options={
                "base_manager_name": "all_objects",
                "default_manager_name": "all_objects",
                "verbose_name": "user profile",
                "verbose_name_plural": "user profiles",
            },
        ),
        migrations.AlterModelManagers(
            name="event",
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="userprofile",
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name="userprofile",
            name="is_visible",
            field=models.BooleanField(default=True, verbose_name="Visible"),
        ),
        migrations.AlterField(
            model_name="qualificationgrant",
            name="expires",
            field=ephios.core.models.users.ExpirationDateField(
                blank=True, null=True, verbose_name="expiration date"
            ),
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-03 15:34

from django.db import migrations

import ephios.core.models.users


class Migration(migrations.Migration):

    dependencies = [
        ("eventautoqualification", "0003_alter_eventautoqualificationconfiguration_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="eventautoqualificationconfiguration",
            name="expiration_date",
            field=ephios.core.models.users.ExpirationDateField(
                blank=True, null=True, verbose_name="Expiration date"
            ),
        ),
    ]



from django.dispatch import receiver

from ephios.core.plugins import PluginSignal
from ephios.core.services.notifications.backends import send_all_notifications
from ephios.core.services.participation import send_participation_finished

register_consequence_handlers = PluginSignal()
"""
This signal is sent out to get all known consequence handlers. Receivers should return a list of
subclasses of ``ephios.core.consequences.BaseConsequenceHandler``.
"""

register_signup_methods = PluginSignal()
"""
This signal is sent out to get all known signup methods. Receivers should return a list of
subclasses of ``ephios.core.signup.methods.BaseSignupMethod``.
"""

check_participant_signup = PluginSignal()
"""
This signal is sent out so receivers can prevent signup for a shift or provide feedback for dispatchers.
Receivers will receive a ``participant`` and ``method`` keyword argument and
should return an instance of the fitting! subclass of ``ephios.core.signup.methods.ParticipationError`` or None.
"""

footer_link = PluginSignal()
"""
This signal is sent out to get links for that page footer. Receivers should return a dict of
with keys being the text and values being the url to link to.
Receivers will receive a ``request`` keyword argument.
"""

administration_settings_section = PluginSignal()
"""
This signal is sent out to get sections for administration settings. Receivers should return a list of dicts
containing key-value-pairs for 'label', 'url' and a boolean flag 'active'.
Receivers will receive a ``request`` keyword argument.
"""

participant_from_request = PluginSignal()
"""
This signal is sent out to get a participant from a request with an unauthenticated user.
Return a subclass of AbstractParticipant or None if you cannot provide a participant.
The first non-None return-value will be used.
Receivers will receive a ``request`` keyword argument.
"""

event_forms = PluginSignal()
"""
This signal is sent out to get a list of form instances to show on the event create and update views.
You receive an `event` and `request` keyword arg you should use to create an instance of your form.
Subclass `BaseEventPluginForm` to customize the rendering behavior.
If all forms are valid, `save` will be called on your form.
"""

register_notification_types = PluginSignal()
"""
This signal is sent out to get all notification types that can be sent out to a user or participant.
Receivers should return a list of subclasses of ``ephios.core.notifications.types.AbstractNotificationHandler``
"""

register_notification_backends = PluginSignal()
"""
This signal is sent out to get all backends that can handle sending out notifications.
Receivers should return a list of subclasses of ``ephios.core.notifications.backends.AbstractNotificationBackend``
"""

periodic_signal = PluginSignal()
"""
This signal is called periodically, at least every 15 minutes.
"""

participation_finished = PluginSignal()
"""
This signal is sent out once for every confirmed participation after the participation ended and
the ``finished`` flag is set to True. Changing the shift date to the future will not cause this to be called again.
Exceptions in receivers are ignored, so make sure your code is robust enough to handle errors and retry if needed.
This signal is based on ``periodic_signal`` and provides a ``participation`` keyword argument.
"""

register_event_bulk_action = PluginSignal()
"""
This signal is sent out to get a list of actions that a user can perform on a list of events.
Receivers should return a list of actions. Each action is represented by a dict with the keys ``url``, ``label`` and ``icon``.
Once the user wants to perform the action, a POST request will be issued to this URL. The ``bulk_action`` field
will contain a list of event ids on which the action should be performed.
"""


@receiver(
    register_consequence_handlers,
    dispatch_uid="ephios.core.signals.register_base_consequence_handlers",
)
def register_base_consequence_handlers(sender, **kwargs):
    from ephios.core.consequences import (
        QualificationConsequenceHandler,
        WorkingHoursConsequenceHandler,
    )

    return [WorkingHoursConsequenceHandler, QualificationConsequenceHandler]


@receiver(
    register_notification_types, dispatch_uid="ephios.core.signals.register_core_notification_types"
)
def register_core_notification_types(sender, **kwargs):
    from ephios.core.services.notifications.types import CORE_NOTIFICATION_TYPES

    return CORE_NOTIFICATION_TYPES


@receiver(
    register_notification_backends,
    dispatch_uid="ephios.core.signals.register_core_notification_backends",
)
def register_core_notification_backends(sender, **kwargs):
    from ephios.core.services.notifications.backends import CORE_NOTIFICATION_BACKENDS

    return CORE_NOTIFICATION_BACKENDS


@receiver(periodic_signal, dispatch_uid="ephios.core.signals.send_notifications")
def send_notifications(sender, **kwargs):
    send_all_notifications()


periodic_signal.connect(
    send_participation_finished, dispatch_uid="ephios.core.signals.send_participation_finished"
)



import copy
import os
from email.utils import getaddresses

try:
    import importlib_metadata  # importlib is broken on python3.8, using backport
except ImportError:
    import importlib.metadata as importlib_metadata

import environ
from cryptography.hazmat.primitives import serialization
from django.contrib.messages import constants
from py_vapid import Vapid, b64urlencode

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

env = environ.Env()
# for syntax see https://django-environ.readthedocs.io/en/latest/
environ.Env.read_env(env_file=os.path.join(BASE_DIR, ".env"))

DATA_DIR = env.str("DATA_DIR", default=os.path.join(BASE_DIR, "data"))
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

SECRET_KEY = env.str("SECRET_KEY")
DEBUG = env.bool("DEBUG")
ALLOWED_HOSTS = env.list("ALLOWED_HOSTS")

if not DEBUG:
    SESSION_COOKIE_SECURE = True
    CSRF_COOKIE_SECURE = True
    X_FRAME_OPTIONS = "DENY"
    SECURE_CONTENT_TYPE_NOSNIFF = True
    SECURE_BROWSER_XSS_FILTER = True
    SECURE_SSL_REDIRECT = True
    SECURE_HSTS_SECONDS = 3600
    SECURE_HSTS_INCLUDE_SUBDOMAINS = True
    SECURE_REFERRER_POLICY = "same-origin"

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "polymorphic",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "guardian",
    "django_select2",
    "djangoformsetjs",
    "compressor",
    "recurrence",
    "statici18n",
    "dynamic_preferences.users.apps.UserPreferencesConfig",
    "crispy_forms",
    "crispy_bootstrap5",
    "webpush",
    "ephios.modellogging",
]

EPHIOS_CORE_MODULES = [
    "ephios.core",
    "ephios.extra",
]
INSTALLED_APPS += EPHIOS_CORE_MODULES

CORE_PLUGINS = [
    "ephios.plugins.basesignup.apps.PluginApp",
    "ephios.plugins.pages.apps.PluginApp",
    "ephios.plugins.qualification_management.apps.PluginApp",
    "ephios.plugins.guests.apps.PluginApp",
    "ephios.plugins.eventautoqualification.apps.PluginApp",
]
PLUGINS = copy.copy(CORE_PLUGINS)
for ep in importlib_metadata.entry_points().get("ephios.plugins", []):
    PLUGINS.append(ep.module)

INSTALLED_APPS += PLUGINS

INSTALLED_APPS += ["dynamic_preferences"]  # must come after our apps to collect preferences

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "csp.middleware.CSPMiddleware",
    "ephios.modellogging.middleware.LoggingRequestMiddleware",
]

ROOT_URLCONF = "ephios.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "ephios/templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
                "dynamic_preferences.processors.global_preferences",
                "ephios.core.context.ephios_base_context",
            ],
        },
    },
]

LOCALE_PATHS = (os.path.join(BASE_DIR, "ephios/locale"),)

WSGI_APPLICATION = "ephios.wsgi.application"

# Database
# https://docs.djangoproject.com/en/3.0/ref/settings/#databases

DATABASES = {"default": env.db_url()}

# Caches
CACHES = {"default": env.cache_url(default="locmemcache://")}
SESSION_ENGINE = "django.contrib.sessions.backends.cached_db"
DJANGO_REDIS_IGNORE_EXCEPTIONS = True
DJANGO_REDIS_LOG_IGNORED_EXCEPTIONS = True

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "ephios.core.services.password_reset.CustomMinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "guardian.backends.ObjectPermissionBackend",
]

AUTH_USER_MODEL = "core.UserProfile"
LOGIN_REDIRECT_URL = "/"
PASSWORD_RESET_TIMEOUT = 28 * 24 * 60 * 60  # seconds

# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = "de"

TIME_ZONE = "Europe/Berlin"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = env.str("STATIC_URL")
STATIC_ROOT = env.str("STATIC_ROOT")
STATICFILES_DIRS = (os.path.join(BASE_DIR, "ephios/static"),)
STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "compressor.finders.CompressorFinder",
)
COMPRESS_ENABLED = not DEBUG
# https://www.accordbox.com/blog/how-use-scss-sass-your-django-project-python-way/
COMPRESS_PRECOMPILERS = (("text/x-scss", "django_libsass.SassCompiler"),)

# mail configuration
EMAIL_CONFIG = env.email_url("EMAIL_URL")
vars().update(EMAIL_CONFIG)
DEFAULT_FROM_EMAIL = env.str("DEFAULT_FROM_EMAIL")
SERVER_EMAIL = env.str("SERVER_EMAIL")
ADMINS = getaddresses([env("ADMINS")])

# logging
if not DEBUG:
    LOGGING = {
        "version": 1,
        "disable_existing_loggers": True,
        "handlers": {
            "mail_admins": {"level": "ERROR", "class": "django.utils.log.AdminEmailHandler"}
        },
        "loggers": {
            "django": {
                "handlers": ["mail_admins"],
                "level": "ERROR",
                "propagate": False,
            },
        },
        "root": {
            "handlers": ["mail_admins"],
            "level": "ERROR",
        },
    }


# Guardian configuration
ANONYMOUS_USER_NAME = None
GUARDIAN_MONKEY_PATCH = False

# django-select2
# Prevent django-select from loading the select2 resources as we want to serve them locally
SELECT2_JS = ""
SELECT2_CSS = ""
SELECT2_I18N_PATH = ""

# django-debug-toolbar
if DEBUG and env.bool("DEBUG_TOOLBAR", True):
    INSTALLED_APPS.append("django_extensions")
    INSTALLED_APPS.append("debug_toolbar")
    MIDDLEWARE.insert(0, "debug_toolbar.middleware.DebugToolbarMiddleware")
    INTERNAL_IPS = env.str("INTERNAL_IPS")

# django-csp
# Bootstrap requires embedded SVG files loaded via a data URI. This is not ideal, but will only be fixed in
# bootstrap v5 or v6. See https://github.com/twbs/bootstrap/issues/25394 for details on the problem and
# https://security.stackexchange.com/a/167244 on why allowing data: is considered okay
CSP_IMG_SRC = ("'self'", "data:")
CSP_STYLE_SRC = ("'self'",)
CSP_INCLUDE_NONCE_IN = ["style-src"]

# django-crispy-forms
CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap5"
CRISPY_TEMPLATE_PACK = "bootstrap5"
CRISPY_FAIL_SILENTLY = not DEBUG

# django.contrib.messages
MESSAGE_TAGS = {
    constants.DEBUG: "alert-info",
    constants.INFO: "alert-info",
    constants.SUCCESS: "alert-success",
    constants.WARNING: "alert-warning",
    constants.ERROR: "alert-danger",
}

# PWA
PWA_APP_ICONS = [
    {"src": "/static/ephios/img/ephios-192x.png", "sizes": "192x192", "purpose": "any maskable"},
    {"src": "/static/ephios/img/ephios-512x.png", "sizes": "512x512", "purpose": "any maskable"},
    {"src": "/static/ephios/img/ephios-1024x.png", "sizes": "1024x1024", "purpose": "any maskable"},
]

# django-webpush
if vapid_private_key_path := env.str("VAPID_PRIVATE_KEY_PATH", None):
    vp = Vapid().from_file(vapid_private_key_path)
    WEBPUSH_SETTINGS = {
        "VAPID_PUBLIC_KEY": b64urlencode(
            vp.public_key.public_bytes(
                serialization.Encoding.X962, serialization.PublicFormat.UncompressedPoint
            )
        ),
        "VAPID_PRIVATE_KEY": vp,
        "VAPID_ADMIN_EMAIL": ADMINS[0][1],
    }


def GET_SITE_URL():
    site_url = env.str("SITE_URL")
    if site_url.endswith("/"):
        site_url = site_url[:-1]
    return site_url



# inspired by https://jfelix.info/blog/how-to-make-a-text-color-fit-any-background-color
from math import sqrt

from ephios.core.models import EventType


def calculate_luminance(rgb: tuple):
    r, g, b = map(
        lambda channel: channel / 255 / 12.92
        if channel / 255 <= 0.03928
        else ((channel / 255 + 0.055) / 1.055) ** 2.4,
        rgb,
    )
    return 0.2126 * r + 0.7152 * g + 0.0722 * b


def get_text_color_for_background(background_luminance: int):
    return "#000000" if background_luminance > sqrt(1.05 * 0.05) - 0.05 else "#ffffff"


def get_eventtype_color_style(eventtype: EventType):
    luminance = calculate_luminance(
        (
            int(eventtype.color[1:3], 16),
            int(eventtype.color[3:5], 16),
            int(eventtype.color[5:7], 16),
        )
    )
    text_color = get_text_color_for_background(luminance)
    return f".badge-{eventtype.pk}-color{{background-color:{eventtype.color};color:{text_color}}}"



from django import forms
from django.contrib import messages
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ephios.core.forms.events import BaseEventPluginFormMixin
from ephios.plugins.guests.models import EventGuestShare


class EventAllowGuestsForm(BaseEventPluginFormMixin, forms.Form):
    active = forms.BooleanField(label=_("Allow guests"), required=False)
    link = forms.CharField(disabled=True, label="Link for guest registration", required=False)
    new_link = forms.BooleanField(label=_("Generate a new link when saving"), required=False)

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("prefix", "guests")
        self.event = kwargs.pop("event")
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)
        try:
            self.instance = EventGuestShare.objects.get(event=self.event)
            self.fields["link"].initial = self.instance.url
        except EventGuestShare.DoesNotExist:
            self.instance = EventGuestShare(event=self.event)
            del self.fields["link"]
            del self.fields["new_link"]
        self.fields["active"].initial = self.instance.active

    def save(self):
        if self.cleaned_data.get("new_link"):
            self.instance.new_token()
        self.instance.active = self.cleaned_data["active"]
        if self.cleaned_data["active"]:
            messages.info(
                self.request,
                mark_safe(
                    _("Guests can sign up for this event <a href={href}>here</a>. ").format(
                        href=self.instance.url
                    )
                ),
            )
        self.instance.save()

    @property
    def heading(self):
        return _("Guests")

    def is_function_active(self):
        return self.instance.active



# Generated by Django 3.1.7 on 2021-03-24 20:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0007_eventtype_color"),
    ]

    operations = [
        migrations.AddField(
            model_name="abstractparticipation",
            name="finished",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="abstractparticipation",
            name="data",
            field=models.JSONField(default=dict, verbose_name="Signup data"),
        ),
    ]



# Generated by Django 3.1.7 on 2021-04-03 21:48

import django.db.models.deletion
from django.db import migrations, models

import ephios.modellogging.json


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("modellogging", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="logentry",
            options={
                "ordering": ("-datetime", "-id"),
                "verbose_name": "Log entry",
                "verbose_name_plural": "Log entries",
            },
        ),
        migrations.AlterField(
            model_name="logentry",
            name="attached_to_object_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="associated_logentries",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="logentry",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="logentries",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="logentry",
            name="data",
            field=models.JSONField(
                decoder=ephios.modellogging.json.LogJSONDecoder,
                default=dict,
                encoder=ephios.modellogging.json.LogJSONEncoder,
            ),
        ),
    ]



import functools
import logging

from django.apps import AppConfig, apps
from django.conf import settings
from django.dispatch import Signal, receiver
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.signals import preference_updated

logger = logging.getLogger(__name__)

# The plugin mechanics are heavily inspired by pretix (licenced under Apache 2.0) - Check it out!

logger.info("Installed plugins: %s", ", ".join(settings.PLUGINS))


def get_all_plugins():
    """
    Return the EphiosPluginMeta classes of all plugins found in the installed Django apps.
    """
    plugins = []
    for app in apps.get_app_configs():
        if hasattr(app, "EphiosPluginMeta"):
            meta = app.EphiosPluginMeta
            meta.module = app.name
            meta.app = app
            plugins.append(meta)
    return sorted(
        plugins,
        key=lambda m: (
            0 if m.module.startswith("ephios.") else 1,
            str(m.name).lower().replace("ephios ", ""),
        ),
    )


def get_enabled_plugins():
    """
    Return a subset of all plugin meta classes - those that are enabled
    """
    enabled_plugins = global_preferences_registry.manager().get("general__enabled_plugins")
    yield from (
        plugin
        for plugin in get_all_plugins()
        if plugin.module in enabled_plugins or not getattr(plugin, "visible", True)
    )


@functools.lru_cache()
def is_receiver_path_enabled(searchpath):
    """
    Return True only if ``searchpath`` (e.g. 'ephios.plugins.basesignup.signals')
    relies in a module that is either an enabled plugin or considered ephios core.
    Uses a cache that gets reset when enabled plugins preference changes.
    """
    enabled_paths = settings.EPHIOS_CORE_MODULES + [
        plugin.module for plugin in get_enabled_plugins()
    ]
    # Not using `startwith`, as we don't want to match "ephios_foobar" against "ephios_foo"
    while True:
        if searchpath in enabled_paths:
            return True
        if len(split := searchpath.rsplit(".", 1)) > 1:
            searchpath, _ = split
        else:
            return False


@receiver(preference_updated, dispatch_uid="ephios.core.plugins.clear_receiver_path_cache")
def clear_receiver_path_cache(sender, **kwargs):
    from ephios.core.dynamic_preferences_registry import EnabledPlugins

    if kwargs.get("name") == EnabledPlugins.name:
        logger.debug(
            "Resetting plugin path cache. Now enabled: %s",
            ", ".join(str(plugin.name) for plugin in get_enabled_plugins()),
        )
        is_receiver_path_enabled.cache_clear()


class PluginSignal(Signal):
    """
    Signal that will only be send out to enabled plugins and ephios core.
    """

    def _live_receivers(self, sender):
        return filter(
            lambda rcv: is_receiver_path_enabled(rcv.__module__), super()._live_receivers(sender)
        )

    def send_to_all_plugins(self, sender, **named):
        return [
            (receiver, receiver(signal=self, sender=sender, **named))
            for receiver in super()._live_receivers(sender)
        ]


class PluginConfig(AppConfig):
    """Superclass for Plugin App Configs. Might use this in the future to implement new features."""



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Prefetch
from django.shortcuts import get_object_or_404
from django_ical.views import ICalFeed
from guardian.shortcuts import get_users_with_perms
from icalendar import vCalAddress

from ephios.core.models import AbstractParticipation, Shift


class EventFeed(ICalFeed):
    file_name = "events.ics"
    timezone = "UTC"
    product_id = "-//ephios//ephios//EN"

    def items(self):
        return Shift.objects.all().filter(event__active=True).order_by("meeting_time")

    def item_title(self, item):
        return item.event.title

    def item_description(self, item):
        return item.event.description

    def item_start_datetime(self, item):
        return item.meeting_time

    def item_end_datetime(self, item):
        return item.end_time

    def item_link(self, item):
        return item.event.get_absolute_url()

    def item_location(self, item):
        return item.event.location

    def item_guid(self, item):
        return f"{item.pk}@{settings.GET_SITE_URL()}"

    def item_organizer(self, item):
        user = get_users_with_perms(item.event, only_with_perms_in=["change_event"]).first()
        email = user.email if user else ""
        return vCalAddress(f"MAILTO:{email}")


class UserEventFeed(EventFeed):
    def __init__(self, user):
        super().__init__()
        self.user = user

    def item_start_datetime(self, item):
        return item.participations.all()[0].start_time

    def item_end_datetime(self, item):
        return item.participations.all()[0].end_time

    def items(self):
        shift_ids = self.user.participations.filter(
            state=AbstractParticipation.States.CONFIRMED
        ).values_list("shift", flat=True)
        return (
            Shift.objects.filter(pk__in=shift_ids)
            .select_related("event")
            .prefetch_related(Prefetch("participations", queryset=self.user.participations.all()))
        )


def user_event_feed_view(request, *args, **kwargs):
    feed = UserEventFeed(get_object_or_404(get_user_model(), **kwargs))
    return feed(request, *args, **kwargs)



import functools

from django.conf import settings
from django.contrib.staticfiles import finders
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.utils.translation import get_language


def manifest(request):
    manifest_json = {
        "name": "ephios",
        "short_name": "ephios",
        "description": "ephios manages events for medical services",
        "start_url": "/",
        "display": "standalone",
        "scope": "/",
        "orientation": "any",
        "background_color": "#fff",
        "theme_color": "#000",
        "status_bar": "default",
        "dir": "auto",
        "icons": settings.PWA_APP_ICONS,
        "lang": get_language(),
    }
    response = JsonResponse(manifest_json)
    response["Service-Worker-Allowed"] = "/"
    return response


@functools.lru_cache
def serviceworker_js():
    with open(finders.find("ephios/js/serviceworker.js"), "rb") as sw_js:
        return sw_js.read()


def serviceworker(request):
    return HttpResponse(
        serviceworker_js(),
        content_type="application/javascript",
    )


def offline(request):
    return render(request, "offline.html")
